""" GGUSTicketsCommand

  The GGUSTickets_Command class is a command class to know about
  the number of active present opened tickets.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from six.moves.urllib_error import URLError

from DIRAC import S_ERROR, S_OK
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getGOCSiteName, getGOCSites
from DIRAC.Core.LCG.GGUSTicketsClient import GGUSTicketsClient
from DIRAC.ResourceStatusSystem.Client.ResourceManagementClient import ResourceManagementClient
from DIRAC.ResourceStatusSystem.Command.Command import Command


class GGUSTicketsCommand(Command):
    """
    GGUSTickets "master" Command
    """

    def __init__(self, args=None, clients=None):

        super(GGUSTicketsCommand, self).__init__(args, clients)

        if "GGUSTicketsClient" in self.apis:
            self.gClient = self.apis["GGUSTicketsClient"]
        else:
            self.gClient = GGUSTicketsClient()

        if "ResourceManagementClient" in self.apis:
            self.rmClient = self.apis["ResourceManagementClient"]
        else:
            self.rmClient = ResourceManagementClient()

    def _storeCommand(self, result):
        """
        Stores the results of doNew method on the database.
        """

        for ggus in result:
            resQuery = self.rmClient.addOrModifyGGUSTicketsCache(
                ggus["GocSite"], ggus["Link"], ggus["OpenTickets"], ggus["Tickets"]
            )
            if not resQuery["OK"]:
                return resQuery
        return S_OK()

    def _prepareCommand(self):
        """
        GGUSTicketsCommand requires one arguments:
        - elementName : <str>

        GGUSTickets are associated with gocDB names, so we have to transform the
        diracSiteName into a gocSiteName.
        """

        if "name" not in self.args:
            return S_ERROR('"name" not found in self.args')
        name = self.args["name"]

        return getGOCSiteName(name)

    def doNew(self, masterParams=None):
        """
        Gets the parameters to run, either from the master method or from its
        own arguments.

        For every elementName ( cannot process bulk queries.. ) contacts the
        ggus client. The server is not very stable, so we protect against crashes.

        If there are ggus tickets, are recorded and then returned.
        """

        if masterParams is not None:
            gocName = masterParams
            gocNames = [gocName]

        else:
            gocName = self._prepareCommand()
            if not gocName["OK"]:
                return gocName
            gocName = gocName["Value"]
            gocNames = [gocName]

        try:
            results = self.gClient.getTicketsList(gocName)
        except URLError as e:
            return S_ERROR("%s %s" % (gocName, e))

        if not results["OK"]:
            return results
        results = results["Value"]

        uniformResult = []

        for gocSite, ggusResult in results.items():

            if gocSite not in gocNames:
                continue

            ggusDict = {}
            ggusDict["GocSite"] = gocSite
            ggusDict["Link"] = ggusResult["URL"]

            del ggusResult["URL"]

            openTickets = 0
            for priorityDict in ggusResult.values():
                openTickets += len(priorityDict)

            ggusDict["Tickets"] = ggusResult
            ggusDict["OpenTickets"] = openTickets

            uniformResult.append(ggusDict)

        storeRes = self._storeCommand(uniformResult)
        if not storeRes["OK"]:
            return storeRes

        return S_OK(uniformResult)

    def doCache(self):
        """
        Method that reads the cache table and tries to read from it. It will
        return a list of dictionaries if there are results.
        """

        gocName = self._prepareCommand()
        if not gocName["OK"]:
            return gocName
        gocName = gocName["Value"]

        result = self.rmClient.selectGGUSTicketsCache(gocSite=gocName)
        if result["OK"]:
            result = S_OK([dict(zip(result["Columns"], res)) for res in result["Value"]])

        return result

    def doMaster(self):
        """
        Master method, which looks little bit spaguetti code, sorry !
        - It gets all gocSites.

        As there is no bulk query, it compares with what we have on the database.
        It queries a portion of them.
        """

        gocSites = getGOCSites()
        if not gocSites["OK"]:
            return gocSites
        gocSites = gocSites["Value"]

        #    resQuery = self.rmClient.selectGGUSTicketsCache( meta = { 'columns' : [ 'GocSite' ] } )
        #    if not resQuery[ 'OK' ]:
        #      return resQuery
        #    resQuery = [ element[0] for element in resQuery[ 'Value' ] ]
        #
        #    gocNamesToQuery = set( gocSites ).difference( set( resQuery ) )

        self.log.info("Processing %s" % ", ".join(gocSites))

        for gocNameToQuery in gocSites:

            #    if gocNameToQuery is None:
            #      self.metrics[ 'failed' ].append( 'None result' )
            #      continue

            result = self.doNew(gocNameToQuery)

            if not result["OK"]:
                self.metrics["failed"].append(result)

        return S_OK(self.metrics)



########################################################################
# File :    CSGlobals.py
# Author :  Ricardo Graciani
########################################################################
"""
Some Helper functions to retrieve common location from the CS
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import imp
import six
from DIRAC.Core.Utilities.Decorators import deprecated
from DIRAC.Core.Utilities.DIRACSingleton import DIRACSingleton
from DIRAC.Core.Utilities.Extensions import extensionsByPriority


@six.add_metaclass(DIRACSingleton)
class Extensions(object):
    def __init__(self):
        self.__modules = {}
        self.__orderedExtNames = []
        self.__csExt = []

    def __load(self):
        if self.__orderedExtNames:
            return
        for extName in extensionsByPriority():
            try:
                res = imp.find_module(extName)
                if res[0]:
                    res[0].close()
                self.__orderedExtNames.append(extName)
                self.__modules[extName] = res
            except ImportError:
                pass

    def getCSExtensions(self):
        if not self.__csExt:
            if six.PY3:
                exts = extensionsByPriority()
            else:
                from DIRAC.ConfigurationSystem.Client.Config import gConfig

                exts = gConfig.getValue("/DIRAC/Extensions", [])

            self.__csExt = []
            for ext in exts:
                if ext.endswith("DIRAC"):
                    ext = ext[:-5]
                # If the extension is now "" (i.e. vanilla DIRAC), don't include it
                if ext:
                    self.__csExt.append(ext)
        return self.__csExt

    @deprecated("Use DIRAC.Core.Utilities.Extensions.extensionsByPriority instead")
    def getInstalledExtensions(self):
        return extensionsByPriority()

    def getExtensionPath(self, extName):
        self.__load()
        return self.__modules[extName][1]

    def getExtensionData(self, extName):
        self.__load()
        return self.__modules[extName]


def getSetup():
    from DIRAC import gConfig

    return gConfig.getValue("/DIRAC/Setup", "")


def getVO(defaultVO=""):
    """
    Return VO from configuration
    """
    from DIRAC import gConfig

    return gConfig.getValue("/DIRAC/VirtualOrganization", defaultVO)


def getCSExtensions():
    """
    Return list of extensions registered in the CS
    They do not include DIRAC
    """
    return Extensions().getCSExtensions()


@deprecated("Use DIRAC.Core.Utilities.Extensions.extensionsByPriority instead")
def getInstalledExtensions():
    """
    Return list of extensions registered in the CS and available in local installation
    """
    return extensionsByPriority()


def skipCACheck():
    from DIRAC import gConfig

    return gConfig.getValue("/DIRAC/Security/SkipCAChecks", "false").lower() in ("y", "yes", "true")


def useServerCertificate():
    from DIRAC import gConfig

    return gConfig.getValue("/DIRAC/Security/UseServerCertificate", "false").lower() in ("y", "yes", "true")



"""
 Set of utilities to retrieve Information from proxy
"""
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

import six
import base64

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Security.X509Chain import X509Chain  # pylint: disable=import-error
from DIRAC.Core.Security.VOMS import VOMS
from DIRAC.Core.Security import Locations

from DIRAC.ConfigurationSystem.Client.Helpers import Registry

__RCSID__ = "$Id$"


def getProxyInfo(proxy=False, disableVOMS=False):
    """
    :Returns: a dict with all the proxy info:

      * values that will be there always
          * 'chain' : chain object containing the proxy
          * 'subject' : subject of the proxy
          * 'issuer' : issuer of the proxy
          * 'isProxy' : bool
          * 'isLimitedProxy' : bool
          * 'validDN' : Valid DN in DIRAC
          * 'validGroup' : Valid Group in DIRAC
          * 'secondsLeft' : Seconds left
      * values that can be there
          * 'path' : path to the file,
          * 'group' : DIRAC group
          * 'groupProperties' : Properties that apply to the DIRAC Group
          * 'username' : DIRAC username
          * 'identity' : DN that generated the proxy
          * 'hostname' : DIRAC host nickname
          * 'VOMS'

    """
    # Discover proxy location
    proxyLocation = False
    if isinstance(proxy, X509Chain):
        chain = proxy
    else:
        if not proxy:
            proxyLocation = Locations.getProxyLocation()
        elif isinstance(proxy, six.string_types):
            proxyLocation = proxy
        if not proxyLocation:
            return S_ERROR(DErrno.EPROXYFIND)
        chain = X509Chain()
        retVal = chain.loadProxyFromFile(proxyLocation)
        if not retVal["OK"]:
            return S_ERROR(DErrno.EPROXYREAD, "%s: %s " % (proxyLocation, retVal["Message"]))

    retVal = chain.getCredentials()
    if not retVal["OK"]:
        return retVal

    infoDict = retVal["Value"]
    infoDict["chain"] = chain
    if proxyLocation:
        infoDict["path"] = proxyLocation

    if not disableVOMS and chain.isVOMS()["Value"]:
        infoDict["hasVOMS"] = True
        retVal = VOMS().getVOMSAttributes(chain)
        if retVal["OK"]:
            infoDict["VOMS"] = retVal["Value"]
        else:
            infoDict["VOMSError"] = retVal["Message"].strip()
    return S_OK(infoDict)


def getProxyInfoAsString(proxyLoc=False, disableVOMS=False):
    """
    return the info as a printable string
    """
    retVal = getProxyInfo(proxyLoc, disableVOMS)
    if not retVal["OK"]:
        return retVal
    infoDict = retVal["Value"]
    return S_OK(formatProxyInfoAsString(infoDict))


def formatProxyInfoAsString(infoDict):
    """
    convert a proxy infoDict into a string
    """
    leftAlign = 13
    contentList = []
    for field in (
        "subject",
        "issuer",
        "identity",
        "subproxyUser",
        ("secondsLeft", "timeleft"),
        ("group", "DIRAC group"),
        "rfc",
        "path",
        "username",
        ("groupProperties", "properties"),
        ("hasVOMS", "VOMS"),
        ("VOMS", "VOMS fqan"),
        ("VOMSError", "VOMS Error"),
    ):
        if isinstance(field, six.string_types):
            dispField = field
        else:
            dispField = field[1]
            field = field[0]
        if field not in infoDict:
            continue
        if field == "secondsLeft":
            secs = infoDict[field]
            hours = int(secs / 3600)
            secs -= hours * 3600
            mins = int(secs / 60)
            secs -= mins * 60
            value = "%02d:%02d:%02d" % (hours, mins, secs)
        elif field == "groupProperties":
            value = ", ".join(infoDict[field])
        else:
            value = infoDict[field]
        contentList.append("%s: %s" % (dispField.ljust(leftAlign), value))
    return "\n".join(contentList)


def getProxyStepsInfo(chain):
    """
    Extended information of all Steps in the ProxyChain
    Returns a list of dictionary with Info for each Step.
    """
    infoList = []
    nC = chain.getNumCertsInChain()["Value"]
    for i in range(nC):
        cert = chain.getCertInChain(i)["Value"]
        stepInfo = {}
        stepInfo["subject"] = cert.getSubjectDN()["Value"]
        stepInfo["issuer"] = cert.getIssuerDN()["Value"]
        stepInfo["serial"] = cert.getSerialNumber()["Value"]
        stepInfo["not before"] = cert.getNotBeforeDate()["Value"]
        stepInfo["not after"] = cert.getNotAfterDate()["Value"]
        stepInfo["lifetime"] = cert.getRemainingSecs()["Value"]
        stepInfo["extensions"] = cert.getExtensions()["Value"]
        dG = cert.getDIRACGroup(ignoreDefault=True)["Value"]
        if dG:
            stepInfo["group"] = dG
        if cert.hasVOMSExtensions()["Value"]:
            stepInfo["VOMS ext"] = True
        infoList.append(stepInfo)
    return S_OK(infoList)


def formatProxyStepsInfoAsString(infoList):
    """
    Format the List of Proxy steps dictionaries as a printable string.
    """
    contentsList = []
    for i in range(len(infoList)):
        contentsList.append(" + Step %s" % i)
        stepInfo = infoList[i]
        for key in (
            "subject",
            "issuer",
            "serial",
            "not after",
            "not before",
            "group",
            "VOMS ext",
            "lifetime",
            "extensions",
        ):
            if key in stepInfo:
                value = stepInfo[key]
                if key == "serial":
                    try:
                        # b16encode needs a string, while the serial
                        # may be a long
                        value = base64.b16encode("%s" % value)
                    except Exception as e:
                        gLogger.exception("Could not read serial:", lException=e)
                if key == "lifetime":
                    secs = value
                    hours = int(secs / 3600)
                    secs -= hours * 3600
                    mins = int(secs / 60)
                    secs -= mins * 60
                    value = "%02d:%02d:%02d" % (hours, mins, secs)
                if key == "extensions":
                    value = "\n   %s" % "\n   ".join(
                        ["%s = %s" % (extName.strip().rjust(20), extValue.strip()) for extName, extValue in value]
                    )
                contentsList.append("  %s : %s" % (key.ljust(10).capitalize(), value))
    return "\n".join(contentsList)


def getVOfromProxyGroup():
    """
    Return the VO associated to the group in the proxy
    """
    voName = Registry.getVOForGroup("NoneExistingGroup")
    ret = getProxyInfo(disableVOMS=True)
    if not ret["OK"]:
        return S_OK(voName)
    if "group" in ret["Value"]:
        voName = Registry.getVOForGroup(ret["Value"]["group"])
    return S_OK(voName)



""" The JobDB class is a front-end to the main WMS database containing
    job definitions and status information. It is used in most of the WMS
    components


**Configuration Parameters**:

The following options can be set in ``Systems/WorkloadManagement/<Setup>/Databases/JobDB``

* *MaxRescheduling*:     Set the maximum number of times a job can be rescheduled, default *3*.
* *CompressJDLs*:        Enable compression of JDLs when they are stored in the database, default *False*.

"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import six
import base64
import zlib

import operator

__RCSID__ = "$Id$"

from DIRAC.ConfigurationSystem.Client.Config import gConfig
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOForGroup
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSiteTier
from DIRAC.Core.Base.DB import DB
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Utilities.ClassAd.ClassAdLight import ClassAd
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR
from DIRAC.Core.Utilities import Time
from DIRAC.Core.Utilities.DErrno import EWMSSUBM, EWMSJMAN
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.Core.Utilities.Decorators import deprecated
from DIRAC.ResourceStatusSystem.Client.SiteStatus import SiteStatus
from DIRAC.WorkloadManagementSystem.Client.JobState.JobManifest import JobManifest
from DIRAC.WorkloadManagementSystem.Client import JobStatus

#############################################################################
# utility functions


def compressJDL(jdl):
    """Return compressed JDL string."""
    return base64.b64encode(zlib.compress(jdl.encode(), -1)).decode()


def extractJDL(compressedJDL):
    """Return decompressed JDL string."""
    # the starting bracket is guaranteeed by JobManager.submitJob
    # we need the check to be backward compatible
    if isinstance(compressedJDL, bytes):
        if compressedJDL.startswith(b"["):
            return compressedJDL.decode()
    else:
        if compressedJDL.startswith("["):
            return compressedJDL
    return zlib.decompress(base64.b64decode(compressedJDL)).decode()


#############################################################################


class JobDB(DB):
    """Interface to MySQL-based JobDB"""

    def __init__(self):
        """Standard Constructor"""

        DB.__init__(self, "JobDB", "WorkloadManagement/JobDB")

        # data member to check if __init__ went through without error
        self.__initialized = False
        self.maxRescheduling = self.getCSOption("MaxRescheduling", 3)

        # loading the function that will be used to determine the platform (it can be VO specific)
        res = ObjectLoader().loadObject("ConfigurationSystem.Client.Helpers.Resources", "getDIRACPlatform")
        if not res["OK"]:
            self.log.fatal(res["Message"])
        self.getDIRACPlatform = res["Value"]

        self.jobAttributeNames = []

        self.siteClient = SiteStatus()

        result = self.__getAttributeNames()

        if not result["OK"]:
            self.log.fatal("JobDB: Can not retrieve job Attributes")
            return

        self.jdl2DBParameters = ["JobName", "JobType", "JobGroup"]

        self.log.info("MaxReschedule", self.maxRescheduling)
        self.log.info("==================================================")
        self.__initialized = True

    def isValid(self):
        """Check if correctly initialised"""
        return self.__initialized

    def __getAttributeNames(self):
        """get Name of Job Attributes defined in DB
        set self.jobAttributeNames to the list of Names
        return S_OK()
        return S_ERROR upon error
        """

        res = self._query("DESCRIBE Jobs")
        if not res["OK"]:
            return res
        self.jobAttributeNames = [row[0] for row in res["Value"]]

        return S_OK()

    #############################################################################
    @deprecated("Use getJobsAttributes")
    def getAttributesForJobList(self, jobIDList, attrList=None):
        """Get attributes for the jobs in the the jobIDList.
        Returns an S_OK structure with a dictionary of dictionaries as its Value:
        ValueDict[jobID][attribute_name] = attribute_value
        """
        if not jobIDList:
            return S_OK({})
        if attrList:
            missingAttr = [repr(x) for x in attrList if x not in self.jobAttributeNames]
            if missingAttr:
                return S_ERROR("JobDB.getAttributesForJobList: Unknown Attribute(s): %s" % ", ".join(missingAttr))
            attrNames = ",".join(str(x) for x in attrList if x in self.jobAttributeNames)
            attr_tmp_list = attrList
        else:
            attrNames = ",".join(self.jobAttributeNames)
            attr_tmp_list = self.jobAttributeNames
        jobList = ",".join([str(x) for x in jobIDList])

        cmd = "SELECT JobID,%s FROM Jobs WHERE JobID in ( %s )" % (attrNames, jobList)
        res = self._query(cmd)
        if not res["OK"]:
            return res
        try:
            retDict = {}
            for retValues in res["Value"]:
                jobID = retValues[0]
                # Make a dict from the list of attributes names and values
                retDict[int(jobID)] = {k: v.decode() for k, v in zip(attr_tmp_list, retValues[1:])}
            return S_OK(retDict)
        except Exception as e:
            return S_ERROR("JobDB.getAttributesForJobList: Failed\n%s" % repr(e))

    #############################################################################
    def getDistinctJobAttributes(self, attribute, condDict=None, older=None, newer=None, timeStamp="LastUpdateTime"):
        """Get distinct values of the job attribute under specified conditions"""
        return self.getDistinctAttributeValues(
            "Jobs", attribute, condDict=condDict, older=older, newer=newer, timeStamp=timeStamp
        )

    #############################################################################
    def getJobParameters(self, jobID, paramList=None):
        """Get Job Parameters defined for jobID.
        Returns a dictionary with the Job Parameters.
        If parameterList is empty - all the parameters are returned.
        """

        if isinstance(jobID, (six.string_types, six.integer_types)):
            jobID = [jobID]

        jobIDList = []
        for jID in jobID:
            ret = self._escapeString(str(jID))
            if not ret["OK"]:
                return ret
            jobIDList.append(ret["Value"])

        # self.log.debug('JobDB.getParameters: Getting Parameters for jobs %s' % ','.join(jobIDList))

        resultDict = {}
        if paramList:
            if isinstance(paramList, six.string_types):
                paramList = paramList.split(",")
            paramNameList = []
            for pn in paramList:
                ret = self._escapeString(pn)
                if not ret["OK"]:
                    return ret
                paramNameList.append(ret["Value"])
            cmd = "SELECT JobID, Name, Value FROM JobParameters WHERE JobID IN (%s) AND Name IN (%s)" % (
                ",".join(jobIDList),
                ",".join(paramNameList),
            )
            result = self._query(cmd)
            if result["OK"]:
                if result["Value"]:
                    for res_jobID, res_name, res_value in result["Value"]:
                        resultDict.setdefault(int(res_jobID), {})[res_name] = res_value.decode(errors="replace")

                return S_OK(resultDict)  # there's a slim chance that this is an empty dictionary
            else:
                return S_ERROR("JobDB.getJobParameters: failed to retrieve parameters")

        else:
            result = self.getFields("JobParameters", ["JobID", "Name", "Value"], {"JobID": jobID})
            if not result["OK"]:
                return result

            for res_jobID, res_name, res_value in result["Value"]:
                resultDict.setdefault(int(res_jobID), {})[res_name] = res_value.decode(errors="replace")

            return S_OK(resultDict)  # there's a slim chance that this is an empty dictionary

    #############################################################################
    def getAtticJobParameters(self, jobID, paramList=None, rescheduleCounter=-1):
        """Get Attic Job Parameters defined for a job with jobID.
        Returns a dictionary with the Attic Job Parameters per each rescheduling cycle.
        If parameterList is empty - all the parameters are returned.
        If recheduleCounter = -1, all cycles are returned.
        """

        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        jobID = ret["Value"]

        # self.log.debug('JobDB.getAtticJobParameters: Getting Attic Parameters for job %s' % jobID)

        resultDict = {}
        paramCondition = ""
        if paramList:
            paramNameList = []
            for x in paramList:
                ret = self._escapeString(x)
                if not ret["OK"]:
                    return ret
                paramNameList.append(ret["Value"])
            paramNames = ",".join(paramNameList)
            paramCondition = " AND Name in (%s)" % paramNames
        rCounter = ""
        if rescheduleCounter != -1:
            rCounter = " AND RescheduleCycle=%d" % int(rescheduleCounter)
        cmd = "SELECT Name, Value, RescheduleCycle from AtticJobParameters"
        cmd += " WHERE JobID=%s %s %s" % (jobID, paramCondition, rCounter)
        result = self._query(cmd)
        if result["OK"]:
            if result["Value"]:
                for name, value, counter in result["Value"]:
                    resultDict.setdefault(counter, {})[name] = value.decode()

            return S_OK(resultDict)
        else:
            return S_ERROR("JobDB.getAtticJobParameters: failed to retrieve parameters")

    #############################################################################
    # TODO: the following 3 methods can be merged into 1.

    def getJobsAttributes(self, jobIDs, attrList=None):
        """Get all Job(s) Attributes for a given list of jobIDs.
        Return a dictionary with all Job Attributes as value pairs
        """

        if not jobIDs:
            return S_OK({})

        # If no list of attributes is given, return all attributes
        if not attrList:
            attrList = self.jobAttributeNames
        if isinstance(attrList, six.string_types):
            attrList = attrList.replace(" ", "").split(",")
        attrList.sort()

        if isinstance(jobIDs, six.string_types):
            jobIDs = [int(jID) for jID in jobIDs.replace(" ", "").split(",")]
        if isinstance(jobIDs, int):
            jobIDs = [jobIDs]

        attrNameListS = []
        for x in attrList:
            ret = self._escapeString(x)
            if not ret["OK"]:
                return ret
            x = "`" + ret["Value"][1:-1] + "`"
            attrNameListS.append(x)
        attrNames = "JobID," + ",".join(attrNameListS)

        cmd = "SELECT %s FROM Jobs WHERE JobID IN (%s)" % (attrNames, ",".join(str(jobID) for jobID in jobIDs))
        res = self._query(cmd)
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_OK({})

        attributes = {}
        for t_att in res["Value"]:
            jobID = int(t_att[0])
            attributes.setdefault(jobID, {})
            for tx, ax in zip(t_att[1:], attrList):
                attributes[jobID].setdefault(ax, tx)

        return S_OK(attributes)

    #############################################################################
    def getJobAttributes(self, jobID, attrList=None):
        """Get all Job Attributes for a given jobID.
        Return a dictionary with all Job Attributes as value pairs
        """

        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        jobID = ret["Value"]

        # If no list is given, return all attributes
        if not attrList:
            attrList = self.jobAttributeNames

        attrNameList = []
        for x in attrList:
            ret = self._escapeString(x)
            if not ret["OK"]:
                return ret
            x = "`" + ret["Value"][1:-1] + "`"
            attrNameList.append(x)
        attrNames = ",".join(attrNameList)
        # self.log.debug('JobDB.getAllJobAttributes: Getting Attributes for job = %s.' % jobID)

        cmd = "SELECT %s FROM Jobs WHERE JobID=%s" % (attrNames, jobID)
        res = self._query(cmd)
        if not res["OK"]:
            return res

        if not res["Value"]:
            return S_OK({})

        values = res["Value"][0]

        attributes = {}
        for name, value in zip(attrList, values):
            attributes[name] = str(value)

        return S_OK(attributes)

    #############################################################################
    def getJobAttribute(self, jobID, attribute):
        """Get the given attribute of a job specified by its jobID"""

        result = self.getJobAttributes(jobID, [attribute])
        if not result["OK"]:
            return result
        return S_OK(result["Value"].get(attribute))

    #############################################################################
    def getJobParameter(self, jobID, parameter):
        """Get the given parameter of a job specified by its jobID"""

        result = self.getJobParameters(jobID, [parameter])
        if not result["OK"]:
            return result
        return S_OK(result.get("Value", {}).get(jobID, {}).get(parameter))

    #############################################################################
    def getJobOptParameter(self, jobID, parameter):
        """Get optimizer parameters for the given job."""

        result = self.getFields("OptimizerParameters", ["Value"], {"JobID": jobID, "Name": parameter})
        if result["OK"]:
            if result["Value"]:
                return S_OK(result["Value"][0][0])
            return S_ERROR("Parameter not found")

        return S_ERROR("Failed to access database")

    #############################################################################
    def getJobOptParameters(self, jobID, paramList=None):
        """Get optimizer parameters for the given job. If the list of parameter names is
        empty, get all the parameters then
        """

        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        jobID = ret["Value"]

        if paramList:
            paramNameList = []
            for x in paramList:
                ret = self._escapeString(x)
                if not ret["OK"]:
                    return ret
                paramNameList.append(ret["Value"])
            paramNames = ",".join(paramNameList)
            cmd = "SELECT Name, Value from OptimizerParameters WHERE JobID=%s and Name in (%s)" % (jobID, paramNames)
        else:
            cmd = "SELECT Name, Value from OptimizerParameters WHERE JobID=%s" % jobID

        result = self._query(cmd)
        if not result["OK"]:
            return S_ERROR("JobDB.getJobOptParameters: failed to retrieve parameters")
        return S_OK({name: value.decode() for name, value in result.get("Value", {})})

    #############################################################################

    def getInputData(self, jobID):
        """Get input data for the given job"""
        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        jobID = ret["Value"]
        cmd = "SELECT LFN FROM InputData WHERE JobID=%s" % jobID
        res = self._query(cmd)
        if not res["OK"]:
            return res

        inputData = [i[0] for i in res["Value"] if i[0].strip()]
        for index, lfn in enumerate(inputData):
            if lfn.lower().startswith("lfn:"):
                inputData[index] = lfn[4:]

        return S_OK(inputData)

    #############################################################################
    def setInputData(self, jobID, inputData):
        """Inserts input data for the given job"""
        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        jobID = ret["Value"]
        cmd = "DELETE FROM InputData WHERE JobID=%s" % (jobID)
        result = self._update(cmd)
        if not result["OK"]:
            result = S_ERROR("JobDB.setInputData: operation failed.")

        for lfn in inputData:
            # some jobs are setting empty string as InputData
            if not lfn:
                continue
            ret = self._escapeString(lfn.strip())
            if not ret["OK"]:
                return ret
            lfn = ret["Value"]
            cmd = "INSERT INTO InputData (JobID,LFN) VALUES (%s, %s )" % (jobID, lfn)
            res = self._update(cmd)
            if not res["OK"]:
                return res

        return S_OK("Files added")

    #############################################################################
    def setOptimizerChain(self, jobID, optimizerList):
        """Set the optimizer chain for the given job. The 'TaskQueue'
        optimizer should be the last one in the chain, it is added
        if not present in the optimizerList
        """

        optString = ",".join(optimizerList)
        return self.setJobOptParameter(jobID, "OptimizerChain", optString)

    #############################################################################
    def setNextOptimizer(self, jobID, currentOptimizer):
        """Set the job status to be processed by the next optimizer in the
        chain
        """

        result = self.getJobOptParameter(jobID, "OptimizerChain")
        if not result["OK"]:
            return result

        optList = result["Value"].split(",")
        if currentOptimizer not in optList:
            return S_ERROR("Could not find " + currentOptimizer + " in chain")
        try:
            # Append None to get a list of (opt,nextOpt)
            optList.append(None)
            nextOptimizer = None
            for opt, nextOptimizer in zip(optList[:-1], optList[1:]):
                if opt == currentOptimizer:
                    break
            if nextOptimizer is None:
                return S_ERROR("Unexpected end of the Optimizer Chain")
        except ValueError:
            return S_ERROR("The " + currentOptimizer + " not found in the chain")

        result = self.setJobStatus(jobID, status=JobStatus.CHECKING, minorStatus=nextOptimizer)
        if not result["OK"]:
            return result
        return S_OK(nextOptimizer)

    ############################################################################
    def selectJobs(self, condDict, older=None, newer=None, timeStamp="LastUpdateTime", orderAttribute=None, limit=None):
        """Select jobs matching the following conditions:
        - condDict dictionary of required Key = Value pairs;
        - with the last update date older and/or newer than given dates;

        The result is ordered by JobID if requested, the result is limited to a given
        number of jobs if requested.
        """

        # self.log.debug('JobDB.selectJobs: retrieving jobs.')

        res = self.getFields(
            "Jobs",
            ["JobID"],
            condDict=condDict,
            limit=limit,
            older=older,
            newer=newer,
            timeStamp=timeStamp,
            orderAttribute=orderAttribute,
        )

        if not res["OK"]:
            return res

        if not res["Value"]:
            return S_OK([])
        return S_OK([self._to_value(i) for i in res["Value"]])

    #############################################################################
    def setJobAttribute(self, jobID, attrName, attrValue, update=False, myDate=None, force=False):
        """Set an attribute value for job specified by jobID.
        The LastUpdate time stamp is refreshed if explicitly requested

        :param jobID: job ID
        :type jobID: int or str
        :param str attrName: attribute name
        :param str attrValue: attribute value
        :param bool update: optional flag to update the job LastUpdateTime stamp
        :param str myDate: optional time stamp for the LastUpdateTime attribute

        :return: S_OK/S_ERROR
        """

        if attrName not in self.jobAttributeNames:
            return S_ERROR(EWMSJMAN, "Request to set non-existing job attribute")

        if attrName == "Status":
            # Treat this update separately
            res = self.setJobsMajorStatus([jobID], attrValue, force=force)
            if not res["OK"]:
                return res
            if update:
                cmd = "UPDATE Jobs SET LastUpdateTime=UTC_TIMESTAMP() WHERE JobID=%s" % jobID
                if myDate:
                    cmd += " AND LastUpdateTime < %s" % myDate
                return self._update(cmd)
            else:
                return res

        # if we are here it's because we are not updating the status
        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        jobID = ret["Value"]

        ret = self._escapeString(attrValue)
        if not ret["OK"]:
            return ret
        value = ret["Value"]

        if update:
            cmd = "UPDATE Jobs SET %s=%s,LastUpdateTime=UTC_TIMESTAMP() WHERE JobID=%s" % (attrName, value, jobID)
        else:
            cmd = "UPDATE Jobs SET %s=%s WHERE JobID=%s" % (attrName, value, jobID)

        if myDate:
            cmd += " AND LastUpdateTime < %s" % myDate

        return self._update(cmd)

    #############################################################################
    def setJobAttributes(self, jobID, attrNames, attrValues, update=False, myDate=None, force=False):
        """Set one or more attribute values for one or more jobs specified by jobID.
        The LastUpdate time stamp is refreshed if explicitly requested with the update flag

        This method is also used for updating the Status, MinorStatus, ApplicationStatus
        of a job, as self.setJobsStatus also calls this method.
        If the status is already final, we don't update it.

        :param jobID: one or more job IDs
        :type jobID: int or str or list
        :param list attrNames: names of attributes to update
        :param list attrValues: corresponding values of attributes to update
        :param bool update: optional flag to update the job LastUpdateTime stamp
        :param str myDate: optional time stamp for the LastUpdateTime attribute
        :param bool force: force update of Status (override State Machine decision)

        :return: S_OK/S_ERROR
        """

        jobIDList = jobID
        if not isinstance(jobID, (list, tuple)):
            jobIDList = [jobID]

        jIDList = []
        for jID in jobIDList:
            ret = self._escapeString(jID)
            if not ret["OK"]:
                return ret
            jIDList.append(ret["Value"])

        if len(attrNames) != len(attrValues):
            return S_ERROR("JobDB.setAttributes: incompatible Argument length")

        for attrName in attrNames:
            if attrName not in self.jobAttributeNames:
                return S_ERROR(EWMSJMAN, "Request to set non-existing job attribute")

        if "Status" in attrNames:
            # Treat this update separately
            res = self.setJobsMajorStatus(jIDList, attrValues[attrNames.index("Status")], force=force)
            if not res["OK"]:
                return res
            attrValues.pop(attrNames.index("Status"))
            attrNames.remove("Status")

        attr = []

        for name, value in zip(attrNames, attrValues):
            ret = self._escapeString(value)
            if not ret["OK"]:
                return ret
            attr.append("%s=%s" % (name, ret["Value"]))
        if update:
            attr.append("LastUpdateTime=UTC_TIMESTAMP()")
        if not attr:
            return S_ERROR("JobDB.setAttributes: Nothing to do")

        cmd = "UPDATE Jobs SET %s WHERE JobID in ( %s )" % (", ".join(attr), ", ".join(jIDList))

        if myDate:
            cmd += " AND LastUpdateTime < %s" % myDate

        return self._update(cmd)

    def setJobsMajorStatus(self, jIDList, candidateStatus, force=False):
        """
        Sets jobs major status, considering the JobStateMachine result

        :param list jIDList: list of one or more job IDs
        :param str candidateStatus: candidate major Status
        """

        # get the current statuses of the jobs
        res = self.getJobsAttributes(jIDList, ["Status"])
        if not res["OK"]:
            return res
        jIDStatusDict = res["Value"]

        newStatuses = {}
        for jID, jIDStatus in jIDStatusDict.items():
            if force:
                self.log.info("Status update forced", "(%s: %s -> %s)" % (str(jID), jIDStatus, candidateStatus))
                nextState = candidateStatus
            else:
                res = JobStatus.JobsStateMachine(jIDStatus["Status"]).getNextState(candidateStatus)
                if not res["OK"]:
                    return res
                nextState = res["Value"]

                # If the JobsStateMachine does not accept the candidate, add it to separate dictionary
                if candidateStatus != nextState:
                    self.log.error(
                        "Job Status Error",
                        "%s can't move from %s to %s: using %s"
                        % (jID, jIDStatus["Status"], candidateStatus, nextState),
                    )

            newStatuses[jID] = nextState

        cmd = "INSERT INTO Jobs (JobID, Status) VALUES "

        ns = []
        for jID, status in newStatuses.items():
            ret_status = self._escapeString(status)
            if not ret_status["OK"]:
                return ret_status
            status = ret_status["Value"]
            ns.append("(%s, %s)" % (jID, status))
        cmd += ",".join(ns)

        cmd += " ON DUPLICATE KEY UPDATE Status=VALUES(Status)"

        return self._update(cmd)

    def setJobStatus(self, jobID, status="", minorStatus="", applicationStatus=""):
        """Set status of the job specified by its jobID"""
        # Do not update the LastUpdate time stamp if setting the Stalled status
        update_flag = True
        if status == JobStatus.STALLED:
            update_flag = False

        attrNames = []
        attrValues = []
        if status:
            attrNames.append("Status")
            attrValues.append(status)
        if minorStatus:
            attrNames.append("MinorStatus")
            attrValues.append(minorStatus)
        if applicationStatus:
            attrNames.append("ApplicationStatus")
            attrValues.append(applicationStatus[:255])

        result = self.setJobAttributes(jobID, attrNames, attrValues, update=update_flag)
        if not result["OK"]:
            return result

        return S_OK()

    #############################################################################
    def setEndExecTime(self, jobID, endDate=None):
        """Set EndExecTime time stamp"""

        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        jobID = ret["Value"]

        if endDate:
            ret = self._escapeString(endDate)
            if not ret["OK"]:
                return ret
            endDate = ret["Value"]
        else:
            endDate = "UTC_TIMESTAMP()"
        req = "UPDATE Jobs SET EndExecTime=%s WHERE JobID=%s AND EndExecTime IS NULL" % (endDate, jobID)
        return self._update(req)

    #############################################################################
    def setStartExecTime(self, jobID, startDate=None):
        """Set StartExecTime time stamp and HeartBeatTime if not already set"""

        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        jobID = ret["Value"]

        if startDate:
            ret = self._escapeString(startDate)
            if not ret["OK"]:
                return ret
            startDate = ret["Value"]
        else:
            startDate = "UTC_TIMESTAMP()"
        # Set also the HeartBeatTime in case the job gets stuck before sending the first HeartBeat
        req = "UPDATE Jobs SET HeartBeatTime=%s WHERE JobID=%s AND HeartBeatTime IS NULL" % (startDate, jobID)
        ret = self._update(req)
        if not ret["OK"]:
            return ret
        req = "UPDATE Jobs SET StartExecTime=%s WHERE JobID=%s AND StartExecTime IS NULL" % (startDate, jobID)
        return self._update(req)

    #############################################################################
    def setJobParameter(self, jobID, key, value):
        """Set a parameter specified by name,value pair for the job JobID"""

        ret = self._escapeString(key)
        if not ret["OK"]:
            return ret
        e_key = ret["Value"]
        ret = self._escapeString(value)
        if not ret["OK"]:
            return ret
        e_value = ret["Value"]

        cmd = "REPLACE JobParameters (JobID,Name,Value) VALUES (%d,%s,%s)" % (int(jobID), e_key, e_value)
        return self._update(cmd)

    #############################################################################
    def setJobParameters(self, jobID, parameters):
        """Set parameters specified by a list of name/value pairs for the job JobID

        :param int jobID: Job ID
        :param list parameters: list of tuples (name, value) pairs

        :return: S_OK/S_ERROR
        """

        if not parameters:
            return S_OK()

        insertValueList = []
        for name, value in parameters:
            ret = self._escapeString(name)
            if not ret["OK"]:
                return ret
            e_name = ret["Value"]
            ret = self._escapeString(value)
            if not ret["OK"]:
                return ret
            e_value = ret["Value"]
            insertValueList.append("(%s,%s,%s)" % (jobID, e_name, e_value))

        cmd = "REPLACE JobParameters (JobID,Name,Value) VALUES %s" % ", ".join(insertValueList)
        return self._update(cmd)

    #############################################################################
    def setJobOptParameter(self, jobID, name, value):
        """Set an optimzer parameter specified by name,value pair for the job JobID"""
        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        e_jobID = ret["Value"]

        ret = self._escapeString(name)
        if not ret["OK"]:
            return ret
        e_name = ret["Value"]

        cmd = "DELETE FROM OptimizerParameters WHERE JobID=%s AND Name=%s" % (e_jobID, e_name)
        res = self._update(cmd)
        if not res["OK"]:
            return res

        return self.insertFields("OptimizerParameters", ["JobID", "Name", "Value"], [jobID, name, value])

    #############################################################################
    def removeJobOptParameter(self, jobID, name):
        """Remove the specified optimizer parameter for jobID"""
        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        jobID = ret["Value"]
        ret = self._escapeString(name)
        if not ret["OK"]:
            return ret
        name = ret["Value"]

        cmd = "DELETE FROM OptimizerParameters WHERE JobID=%s AND Name=%s" % (jobID, name)
        return self._update(cmd)

    #############################################################################
    def setAtticJobParameter(self, jobID, key, value, rescheduleCounter):
        """Set attic parameter for job specified by its jobID when job rescheduling
        for later debugging
        """
        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        jobID = ret["Value"]

        ret = self._escapeString(key)
        if not ret["OK"]:
            return ret
        key = ret["Value"]

        ret = self._escapeString(value)
        if not ret["OK"]:
            return ret
        value = ret["Value"]

        ret = self._escapeString(rescheduleCounter)
        if not ret["OK"]:
            return ret
        rescheduleCounter = ret["Value"]

        cmd = "INSERT INTO AtticJobParameters (JobID,RescheduleCycle,Name,Value) VALUES(%s,%s,%s,%s)" % (
            jobID,
            rescheduleCounter,
            key,
            value,
        )
        return self._update(cmd)

    #############################################################################
    def __setInitialJobParameters(self, classadJob, jobID):
        """Set initial job parameters as was defined in the Classad"""

        # Extract initital job parameters
        parameters = {}
        if classadJob.lookupAttribute("Parameters"):
            parameters = classadJob.getDictionaryFromSubJDL("Parameters")
        return self.setJobParameters(jobID, list(parameters.items()))

    #############################################################################
    def setJobJDL(self, jobID, jdl=None, originalJDL=None):
        """Insert JDL's for job specified by jobID"""
        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        jobID = ret["Value"]

        req = "SELECT OriginalJDL FROM JobJDLs WHERE JobID=%s" % jobID
        result = self._query(req)
        updateFlag = False
        if result["OK"] and result["Value"]:
            updateFlag = True

        if jdl:
            ret = self._escapeString(compressJDL(jdl))
            if not ret["OK"]:
                return ret
            e_JDL = ret["Value"]

            if updateFlag:
                cmd = "UPDATE JobJDLs Set JDL=%s WHERE JobID=%s" % (e_JDL, jobID)
            else:
                cmd = "INSERT INTO JobJDLs (JobID,JDL) VALUES (%s,%s)" % (jobID, e_JDL)
            result = self._update(cmd)
            if not result["OK"]:
                return result

        if originalJDL:
            ret = self._escapeString(compressJDL(originalJDL))
            if not ret["OK"]:
                return ret
            e_originalJDL = ret["Value"]

            if updateFlag:
                cmd = "UPDATE JobJDLs Set OriginalJDL=%s WHERE JobID=%s" % (e_originalJDL, jobID)
            else:
                cmd = "INSERT INTO JobJDLs (JobID,OriginalJDL) VALUES (%s,%s)" % (jobID, e_originalJDL)

            result = self._update(cmd)

        return result

    #############################################################################
    def __insertNewJDL(self, jdl):
        """Insert a new JDL in the system, this produces a new JobID"""

        err = "JobDB.__insertNewJDL: Failed to retrieve a new Id."

        result = self.insertFields("JobJDLs", ["JDL", "JobRequirements", "OriginalJDL"], ["", "", compressJDL(jdl)])
        if not result["OK"]:
            self.log.error("Can not insert New JDL", result["Message"])
            return result

        if "lastRowId" not in result:
            return S_ERROR("%s" % err)

        jobID = int(result["lastRowId"])

        self.log.info("JobDB: New JobID served", "%s" % jobID)

        return S_OK(jobID)

    #############################################################################
    def getJobJDL(self, jobID, original=False):
        """Get JDL for job specified by its jobID. By default the current job JDL
        is returned. If 'original' argument is True, original JDL is returned
        """
        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        jobID = ret["Value"]

        if original:
            cmd = "SELECT OriginalJDL FROM JobJDLs WHERE JobID=%s" % jobID
        else:
            cmd = "SELECT JDL FROM JobJDLs WHERE JobID=%s" % jobID

        result = self._query(cmd)
        if result["OK"]:
            jdl = result["Value"]
            if not jdl:
                return S_OK(jdl)
            return S_OK(extractJDL(jdl[0][0]))
        return result

    #############################################################################
    def insertNewJobIntoDB(
        self,
        jdl,
        owner,
        ownerDN,
        ownerGroup,
        diracSetup,
        initialStatus=JobStatus.RECEIVED,
        initialMinorStatus="Job accepted",
    ):
        """Insert the initial JDL into the Job database,
        Do initial JDL crosscheck,
        Set Initial job Attributes and Status

        :param str jdl: job description JDL
        :param str owner: job owner user name
        :param str ownerDN: job owner DN
        :param str ownerGroup: job owner group
        :param str diracSetup: setup in which context the job is submitted
        :param str initialStatus: optional initial job status (Received by default)
        :param str initialMinorStatus: optional initial minor job status
        :return: new job ID
        """
        jobManifest = JobManifest()
        result = jobManifest.load(jdl)
        if not result["OK"]:
            return result
        jobManifest.setOptionsFromDict(
            {"OwnerName": owner, "OwnerDN": ownerDN, "OwnerGroup": ownerGroup, "DIRACSetup": diracSetup}
        )
        result = jobManifest.check()
        if not result["OK"]:
            return result
        jobAttrNames = []
        jobAttrValues = []

        # 1.- insert original JDL on DB and get new JobID
        # Fix the possible lack of the brackets in the JDL
        if jdl.strip()[0].find("[") != 0:
            jdl = "[" + jdl + "]"
        result = self.__insertNewJDL(jdl)
        if not result["OK"]:
            return S_ERROR(EWMSSUBM, "Failed to insert JDL in to DB")
        jobID = result["Value"]

        jobManifest.setOption("JobID", jobID)

        jobAttrNames.append("JobID")
        jobAttrValues.append(jobID)

        jobAttrNames.append("LastUpdateTime")
        jobAttrValues.append(Time.toString())

        jobAttrNames.append("SubmissionTime")
        jobAttrValues.append(Time.toString())

        jobAttrNames.append("Owner")
        jobAttrValues.append(owner)

        jobAttrNames.append("OwnerDN")
        jobAttrValues.append(ownerDN)

        jobAttrNames.append("OwnerGroup")
        jobAttrValues.append(ownerGroup)

        jobAttrNames.append("DIRACSetup")
        jobAttrValues.append(diracSetup)

        # 2.- Check JDL and Prepare DIRAC JDL
        jobJDL = jobManifest.dumpAsJDL()

        # Replace the JobID placeholder if any
        if jobJDL.find("%j") != -1:
            jobJDL = jobJDL.replace("%j", str(jobID))

        classAdJob = ClassAd(jobJDL)
        classAdReq = ClassAd("[]")
        retVal = S_OK(jobID)
        retVal["JobID"] = jobID
        if not classAdJob.isOK():
            jobAttrNames.append("Status")
            jobAttrValues.append(JobStatus.FAILED)

            jobAttrNames.append("MinorStatus")
            jobAttrValues.append("Error in JDL syntax")

            result = self.insertFields("Jobs", jobAttrNames, jobAttrValues)
            if not result["OK"]:
                return result

            retVal["Status"] = JobStatus.FAILED
            retVal["MinorStatus"] = "Error in JDL syntax"
            return retVal

        classAdJob.insertAttributeInt("JobID", jobID)
        result = self.__checkAndPrepareJob(
            jobID, classAdJob, classAdReq, owner, ownerDN, ownerGroup, diracSetup, jobAttrNames, jobAttrValues
        )
        if not result["OK"]:
            return result

        priority = classAdJob.getAttributeInt("Priority")
        if priority is None:
            priority = 0
        jobAttrNames.append("UserPriority")
        jobAttrValues.append(priority)

        for jdlName in self.jdl2DBParameters:
            # Defaults are set by the DB.
            jdlValue = classAdJob.getAttributeString(jdlName)
            if jdlValue:
                jobAttrNames.append(jdlName)
                jobAttrValues.append(jdlValue)

        jdlValue = classAdJob.getAttributeString("Site")
        if jdlValue:
            jobAttrNames.append("Site")
            if jdlValue.find(",") != -1:
                jobAttrValues.append("Multiple")
            else:
                jobAttrValues.append(jdlValue)

        jobAttrNames.append("VerifiedFlag")
        jobAttrValues.append("True")

        jobAttrNames.append("Status")
        jobAttrValues.append(initialStatus)

        jobAttrNames.append("MinorStatus")
        jobAttrValues.append(initialMinorStatus)

        reqJDL = classAdReq.asJDL()
        classAdJob.insertAttributeInt("JobRequirements", reqJDL)

        jobJDL = classAdJob.asJDL()

        result = self.setJobJDL(jobID, jobJDL)
        if not result["OK"]:
            return result

        # Adding the job in the Jobs table
        result = self.insertFields("Jobs", jobAttrNames, jobAttrValues)
        if not result["OK"]:
            return result

        # Setting the Job parameters
        result = self.__setInitialJobParameters(classAdJob, jobID)
        if not result["OK"]:
            return result

        # Looking for the Input Data
        inputData = []
        if classAdJob.lookupAttribute("InputData"):
            inputData = classAdJob.getListFromExpression("InputData")
        values = []

        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        e_jobID = ret["Value"]

        for lfn in inputData:
            # some jobs are setting empty string as InputData
            if not lfn:
                continue
            ret = self._escapeString(lfn.strip())
            if not ret["OK"]:
                return ret
            lfn = ret["Value"]

            values.append("(%s, %s )" % (e_jobID, lfn))

        if values:
            cmd = "INSERT INTO InputData (JobID,LFN) VALUES %s" % ", ".join(values)
            result = self._update(cmd)
            if not result["OK"]:
                return result

        retVal["Status"] = initialStatus
        retVal["MinorStatus"] = initialMinorStatus

        return retVal

    def __checkAndPrepareJob(
        self, jobID, classAdJob, classAdReq, owner, ownerDN, ownerGroup, diracSetup, jobAttrNames, jobAttrValues
    ):
        """
        Check Consistency of Submitted JDL and set some defaults
        Prepare subJDL with Job Requirements
        """
        error = ""
        vo = getVOForGroup(ownerGroup)

        jdlDiracSetup = classAdJob.getAttributeString("DIRACSetup")
        jdlOwner = classAdJob.getAttributeString("Owner")
        jdlOwnerDN = classAdJob.getAttributeString("OwnerDN")
        jdlOwnerGroup = classAdJob.getAttributeString("OwnerGroup")
        jdlVO = classAdJob.getAttributeString("VirtualOrganization")

        # The below is commented out since this is always overwritten by the submitter IDs
        # but the check allows to findout inconsistent client environments
        if jdlDiracSetup and jdlDiracSetup != diracSetup:
            error = "Wrong DIRAC Setup in JDL"
        if jdlOwner and jdlOwner != owner:
            error = "Wrong Owner in JDL"
        elif jdlOwnerDN and jdlOwnerDN != ownerDN:
            error = "Wrong Owner DN in JDL"
        elif jdlOwnerGroup and jdlOwnerGroup != ownerGroup:
            error = "Wrong Owner Group in JDL"
        elif jdlVO and jdlVO != vo:
            error = "Wrong Virtual Organization in JDL"

        classAdJob.insertAttributeString("Owner", owner)
        classAdJob.insertAttributeString("OwnerDN", ownerDN)
        classAdJob.insertAttributeString("OwnerGroup", ownerGroup)

        if vo:
            classAdJob.insertAttributeString("VirtualOrganization", vo)

        classAdReq.insertAttributeString("Setup", diracSetup)
        classAdReq.insertAttributeString("OwnerDN", ownerDN)
        classAdReq.insertAttributeString("OwnerGroup", ownerGroup)
        if vo:
            classAdReq.insertAttributeString("VirtualOrganization", vo)

        inputDataPolicy = Operations(vo=vo).getValue("InputDataPolicy/InputDataModule")
        if inputDataPolicy and not classAdJob.lookupAttribute("InputDataModule"):
            classAdJob.insertAttributeString("InputDataModule", inputDataPolicy)

        # ################## adding DIRAC/VOPolicy as classAds
        # FIXME: to remove
        setup = gConfig.getValue("/DIRAC/Setup", "")
        voPolicyDict = gConfig.getOptionsDict("/DIRAC/VOPolicy/%s/%s" % (vo, setup))
        # voPolicyDict = gConfig.getOptionsDict('/DIRAC/VOPolicy')
        if voPolicyDict["OK"]:
            voPolicy = voPolicyDict["Value"]
            for param, val in voPolicy.items():
                if not classAdJob.lookupAttribute(param):
                    classAdJob.insertAttributeString(param, val)
        # ##################

        # priority
        priority = classAdJob.getAttributeInt("Priority")
        if priority is None:
            priority = 0
        classAdReq.insertAttributeInt("UserPriority", priority)

        # CPU time
        cpuTime = classAdJob.getAttributeInt("CPUTime")
        if cpuTime is None:
            opsHelper = Operations(group=ownerGroup, setup=diracSetup)
            cpuTime = opsHelper.getValue("JobDescription/DefaultCPUTime", 86400)
        classAdReq.insertAttributeInt("CPUTime", cpuTime)

        # platform(s)
        platformList = classAdJob.getListFromExpression("Platform")
        if platformList:
            result = self.getDIRACPlatform(platformList)
            if not result["OK"]:
                return result
            if result["Value"]:
                classAdReq.insertAttributeVectorString("Platforms", result["Value"])
            else:
                error = "OS compatibility info not found"

        if error:
            retVal = S_ERROR(EWMSSUBM, error)
            retVal["JobId"] = jobID
            retVal["Status"] = JobStatus.FAILED
            retVal["MinorStatus"] = error

            jobAttrNames.append("Status")
            jobAttrValues.append(JobStatus.FAILED)

            jobAttrNames.append("MinorStatus")
            jobAttrValues.append(error)
            resultInsert = self.setJobAttributes(jobID, jobAttrNames, jobAttrValues)
            if not resultInsert["OK"]:
                retVal["MinorStatus"] += "; %s" % resultInsert["Message"]

            return retVal

        return S_OK()

    #############################################################################
    def removeJobFromDB(self, jobIDs):
        """
        Remove jobs from the Job DB and clean up all the job related data in various tables
        """

        # ret = self._escapeString(jobID)
        # if not ret['OK']:
        #  return ret
        # e_jobID = ret['Value']

        if not jobIDs:
            return S_OK()

        if not isinstance(jobIDs, list):
            jobIDList = [jobIDs]
        else:
            jobIDList = jobIDs

        failedTablesList = []
        for table in [
            "InputData",
            "JobParameters",
            "AtticJobParameters",
            "HeartBeatLoggingInfo",
            "OptimizerParameters",
            "JobCommands",
            "Jobs",
            "JobJDLs",
        ]:

            cmd = "DELETE FROM %s WHERE JobID in (%s)" % (table, ",".join(str(j) for j in jobIDList))
            result = self._update(cmd)
            if not result["OK"]:
                failedTablesList.append(table)

        result = S_OK()
        if failedTablesList:
            result = S_ERROR("Errors while job removal (tables %s)" % ",".join(failedTablesList))

        return result

    #################################################################
    def rescheduleJobs(self, jobIDs):
        """Reschedule all the jobs in the given list"""

        result = S_OK()

        failedJobs = []
        for jobID in jobIDs:
            result = self.rescheduleJob(jobID)
            if not result["OK"]:
                failedJobs.append(jobID)

        if failedJobs:
            result = S_ERROR("JobDB.rescheduleJobs: Not all the jobs were rescheduled")
            result["FailedJobs"] = failedJobs

        return result

    #############################################################################
    def rescheduleJob(self, jobID):
        """Reschedule the given job to run again from scratch. Retain the already
        defined parameters in the parameter Attic
        """
        # Check Verified Flag
        result = self.getJobAttributes(
            jobID,
            [
                "Status",
                "MinorStatus",
                "VerifiedFlag",
                "RescheduleCounter",
                "Owner",
                "OwnerDN",
                "OwnerGroup",
                "DIRACSetup",
            ],
        )
        if result["OK"]:
            resultDict = result["Value"]
        else:
            return S_ERROR("JobDB.getJobAttributes: can not retrieve job attributes")

        if "VerifiedFlag" not in resultDict:
            return S_ERROR("Job " + str(jobID) + " not found in the system")

        if not resultDict["VerifiedFlag"]:
            return S_ERROR(
                "Job %s not Verified: Status = %s, MinorStatus = %s"
                % (jobID, resultDict["Status"], resultDict["MinorStatus"])
            )

        # Check the Reschedule counter first
        rescheduleCounter = int(resultDict["RescheduleCounter"]) + 1

        self.maxRescheduling = self.getCSOption("MaxRescheduling", self.maxRescheduling)

        # Exit if the limit of the reschedulings is reached
        if rescheduleCounter > self.maxRescheduling:
            self.log.warn("Maximum number of reschedulings is reached", "Job %s" % jobID)
            res = self.setJobStatus(jobID, status=JobStatus.FAILED, minorStatus="Maximum of reschedulings reached")
            if not res["OK"]:
                return res
            return S_ERROR("Maximum number of reschedulings is reached: %s" % self.maxRescheduling)

        jobAttrNames = []
        jobAttrValues = []

        jobAttrNames.append("RescheduleCounter")
        jobAttrValues.append(rescheduleCounter)

        # Save the job parameters for later debugging
        result = self.getJobParameters(jobID)
        if result["OK"]:
            parDict = result["Value"]
            for key, value in parDict.get(jobID, {}).items():
                result = self.setAtticJobParameter(jobID, key, value, rescheduleCounter - 1)
                if not result["OK"]:
                    break

        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        e_jobID = ret["Value"]

        cmd = "DELETE FROM JobParameters WHERE JobID=%s" % e_jobID
        res = self._update(cmd)
        if not res["OK"]:
            return res

        # Delete optimizer parameters
        cmd = "DELETE FROM OptimizerParameters WHERE JobID=%s" % (e_jobID)
        if not self._update(cmd)["OK"]:
            return S_ERROR("JobDB.removeJobOptParameter: operation failed.")

        # the JobManager needs to know if there is InputData ??? to decide which optimizer to call
        # proposal: - use the getInputData method
        res = self.getJobJDL(jobID, original=True)
        if not res["OK"]:
            return res

        jdl = res["Value"]
        # Fix the possible lack of the brackets in the JDL
        if jdl.strip()[0].find("[") != 0:
            jdl = "[" + jdl + "]"
        classAdJob = ClassAd(jdl)
        classAdReq = ClassAd("[]")
        retVal = S_OK(jobID)
        retVal["JobID"] = jobID

        classAdJob.insertAttributeInt("JobID", jobID)
        result = self.__checkAndPrepareJob(
            jobID,
            classAdJob,
            classAdReq,
            resultDict["Owner"],
            resultDict["OwnerDN"],
            resultDict["OwnerGroup"],
            resultDict["DIRACSetup"],
            jobAttrNames,
            jobAttrValues,
        )

        if not result["OK"]:
            return result

        priority = classAdJob.getAttributeInt("Priority")
        if priority is None:
            priority = 0
        jobAttrNames.append("UserPriority")
        jobAttrValues.append(priority)

        siteList = classAdJob.getListFromExpression("Site")
        if not siteList:
            site = "ANY"
        elif len(siteList) > 1:
            site = "Multiple"
        else:
            site = siteList[0]

        jobAttrNames.append("Site")
        jobAttrValues.append(site)

        jobAttrNames.append("Status")
        jobAttrValues.append(JobStatus.RECEIVED)

        jobAttrNames.append("MinorStatus")
        jobAttrValues.append("Job Rescheduled")

        jobAttrNames.append("ApplicationStatus")
        jobAttrValues.append("Unknown")

        jobAttrNames.append("ApplicationNumStatus")
        jobAttrValues.append(0)

        jobAttrNames.append("LastUpdateTime")
        jobAttrValues.append(Time.toString())

        jobAttrNames.append("RescheduleTime")
        jobAttrValues.append(Time.toString())

        reqJDL = classAdReq.asJDL()
        classAdJob.insertAttributeInt("JobRequirements", reqJDL)

        jobJDL = classAdJob.asJDL()

        # Replace the JobID placeholder if any
        if jobJDL.find("%j") != -1:
            jobJDL = jobJDL.replace("%j", str(jobID))

        result = self.setJobJDL(jobID, jobJDL)
        if not result["OK"]:
            return result

        result = self.__setInitialJobParameters(classAdJob, jobID)
        if not result["OK"]:
            return result

        result = self.setJobAttributes(jobID, jobAttrNames, jobAttrValues, force=True)
        if not result["OK"]:
            return result

        retVal["InputData"] = classAdJob.lookupAttribute("InputData")
        retVal["RescheduleCounter"] = rescheduleCounter
        retVal["Status"] = JobStatus.RECEIVED
        retVal["MinorStatus"] = "Job Rescheduled"

        return retVal

    #############################################################################
    def getUserSitesTuple(self, sites):
        """Returns tuple of active/banned/invalid sties from a user provided list."""
        ret = self._escapeValues(sites)
        if not ret["OK"]:
            return ret

        sites = set(sites)
        sitesSql = ret["Value"]
        sitesSql[0] = "SELECT %s AS Site" % sitesSql[0]
        sitesSql = " UNION SELECT ".join(sitesSql)
        cmd = "SELECT Site FROM (%s) " % sitesSql
        cmd += "AS tmptable WHERE Site NOT IN (SELECT Site FROM SiteMask WHERE Status='Active')"
        result = self._query(cmd)
        if not result["OK"]:
            return result
        nonActiveSites = set(x[0] for x in result["Value"])
        activeSites = sites.difference(nonActiveSites)
        bannedSites = nonActiveSites.intersection(set(self.getSiteMask("Banned")))
        invalidSites = nonActiveSites.difference(bannedSites)
        return S_OK((activeSites, bannedSites, invalidSites))

    #############################################################################
    def getSiteMask(self, siteState="Active"):
        """Get the currently active site list"""

        ret = self._escapeString(siteState)
        if not ret["OK"]:
            return ret
        siteState = ret["Value"]

        if siteState == "All":
            cmd = "SELECT Site FROM SiteMask"
        else:
            cmd = "SELECT Site FROM SiteMask WHERE Status=%s" % siteState

        result = self._query(cmd)
        siteList = []
        if result["OK"]:
            siteList = [x[0] for x in result["Value"]]
        else:
            return S_ERROR(DErrno.EMYSQL, "SQL query failed: %s" % cmd)

        return S_OK(siteList)

    #############################################################################
    def getSiteMaskStatus(self, sites=None):
        """Get the current site mask status

        :param sites: A string for a single site to check, or a list to check multiple sites.
        :returns: If input was a list, a dictionary of sites, keys are site
                 names and values are the site statuses. Unknown sites are
                 not included in the output dictionary.
                 If input was a string, then a single value with that site's
                 status, or S_ERROR if the site does not exist in the DB.
        """
        if isinstance(sites, list):
            safeSites = []
            for site in sites:
                res = self._escapeString(site)
                if not res["OK"]:
                    return res
                safeSites.append(res["Value"])
            sitesString = ",".join(safeSites)
            cmd = "SELECT Site, Status FROM SiteMask WHERE Site in (%s)" % sitesString

            result = self._query(cmd)
            return S_OK(dict(result["Value"]))

        elif isinstance(sites, str):

            ret = self._escapeString(sites)
            if not ret["OK"]:
                return ret
            cmd = "SELECT Status FROM SiteMask WHERE Site=%s" % ret["Value"]
            result = self._query(cmd)
            if result["Value"]:
                return S_OK(result["Value"][0][0])
            return S_ERROR("Unknown site %s" % sites)

        else:
            cmd = "SELECT Site,Status FROM SiteMask"

        result = self._query(cmd)
        siteDict = {}
        if result["OK"]:
            for site, status in result["Value"]:
                siteDict[site] = status
        else:
            return S_ERROR(DErrno.EMYSQL, "SQL query failed: %s" % cmd)

        return S_OK(siteDict)

    #############################################################################
    def getAllSiteMaskStatus(self):
        """Get the everything from site mask status"""
        cmd = "SELECT Site,Status,LastUpdateTime,Author,Comment FROM SiteMask"

        result = self._query(cmd)

        if not result["OK"]:
            return result["Message"]

        siteDict = {}
        if result["OK"]:
            for site, status, lastUpdateTime, author, comment in result["Value"]:
                siteDict[site] = status, lastUpdateTime, author, comment

        return S_OK(siteDict)

    #############################################################################
    def setSiteMask(self, siteMaskList, authorDN="Unknown", comment="No comment"):
        """Set the Site Mask to the given mask in a form of a list of tuples (site,status)"""

        for site, status in siteMaskList:
            result = self.__setSiteStatusInMask(site, status, authorDN, comment)
            if not result["OK"]:
                return result

        return S_OK()

    #############################################################################
    def __setSiteStatusInMask(self, site, status, authorDN="Unknown", comment="No comment"):
        """Set the given site status to 'status' or add a new active site"""

        result = self._escapeString(site)
        if not result["OK"]:
            return result
        site = result["Value"]

        result = self._escapeString(status)
        if not result["OK"]:
            return result
        status = result["Value"]

        result = self._escapeString(authorDN)
        if not result["OK"]:
            return result
        authorDN = result["Value"]

        result = self._escapeString(comment)
        if not result["OK"]:
            return result
        comment = result["Value"]

        req = "SELECT Status FROM SiteMask WHERE Site=%s" % site
        result = self._query(req)
        if result["OK"]:
            if result["Value"]:
                current_status = result["Value"][0][0]
                if current_status == status:
                    return S_OK()
                else:
                    req = (
                        "UPDATE SiteMask SET Status=%s,LastUpdateTime=UTC_TIMESTAMP(),"
                        "Author=%s, Comment=%s WHERE Site=%s"
                    )
                    req = req % (status, authorDN, comment, site)
            else:
                req = "INSERT INTO SiteMask VALUES (%s,%s,UTC_TIMESTAMP(),%s,%s)" % (site, status, authorDN, comment)
            result = self._update(req)
            if not result["OK"]:
                return S_ERROR("Failed to update the Site Mask")
            # update the site mask logging record
            req = "INSERT INTO SiteMaskLogging VALUES (%s,%s,UTC_TIMESTAMP(),%s,%s)" % (site, status, authorDN, comment)
            result = self._update(req)
            if not result["OK"]:
                self.log.warn("Failed to update site mask logging", "for %s" % site)
        else:
            return S_ERROR("Failed to get the Site Status from the Mask")

        return S_OK()

    #############################################################################
    def banSiteInMask(self, site, authorDN="Unknown", comment="No comment"):
        """Forbid the given site in the Site Mask"""

        return self.__setSiteStatusInMask(site, "Banned", authorDN, comment)

    #############################################################################
    def allowSiteInMask(self, site, authorDN="Unknown", comment="No comment"):
        """Forbid the given site in the Site Mask"""

        return self.__setSiteStatusInMask(site, "Active", authorDN, comment)

    #############################################################################
    def removeSiteFromMask(self, site=None):
        """Remove the given site from the mask"""
        if not site:
            req = "DELETE FROM SiteMask"
        else:
            ret = self._escapeString(site)
            if not ret["OK"]:
                return ret
            site = ret["Value"]
            req = "DELETE FROM SiteMask WHERE Site=%s" % site

        return self._update(req)

    #############################################################################
    def getSiteMaskLogging(self, siteList):
        """Get the site mask logging history for the list if site names"""

        if siteList:
            siteString = ",".join(["'" + x + "'" for x in siteList])
            req = "SELECT Site,Status,UpdateTime,Author,Comment FROM SiteMaskLogging WHERE Site in (%s)" % siteString
        else:
            req = "SELECT Site,Status,UpdateTime,Author,Comment FROM SiteMaskLogging"
        req += " ORDER BY UpdateTime ASC"

        result = self._query(req)
        if not result["OK"]:
            return result

        availableSiteList = []
        for row in result["Value"]:
            site, status, utime, author, comment = row
            availableSiteList.append(site)

        resultDict = {}
        for site in siteList:
            if not result["Value"] or site not in availableSiteList:
                ret = self._escapeString(site)
                if not ret["OK"]:
                    continue
                e_site = ret["Value"]
                req = "SELECT Status Site,Status,LastUpdateTime,Author,Comment FROM SiteMask WHERE Site=%s" % e_site
                resSite = self._query(req)
                if resSite["OK"]:
                    if resSite["Value"]:
                        site, status, lastUpdate, author, comment = resSite["Value"][0]
                        resultDict[site] = [[status, str(lastUpdate), author, comment]]
                    else:
                        resultDict[site] = [["Unknown", "", "", "Site not present in logging table"]]

        for row in result["Value"]:
            site, status, utime, author, comment = row
            if site not in resultDict:
                resultDict[site] = []
            resultDict[site].append([status, str(utime), author, comment])

        return S_OK(resultDict)

    #############################################################################
    def getSiteSummary(self):
        """Get the summary of jobs in a given status on all the sites"""

        waitingList = ['"Submitted"', '"Assigned"', '"Waiting"', '"Matched"']
        waitingString = ",".join(waitingList)

        result = self.getDistinctJobAttributes("Site")
        if not result["OK"]:
            return result

        siteList = result["Value"]
        siteDict = {}
        totalDict = {
            JobStatus.WAITING: 0,
            JobStatus.RUNNING: 0,
            JobStatus.STALLED: 0,
            JobStatus.DONE: 0,
            JobStatus.FAILED: 0,
        }

        for site in siteList:
            if site == "ANY":
                continue
            # Waiting
            siteDict[site] = {}
            ret = self._escapeString(site)
            if not ret["OK"]:
                return ret
            e_site = ret["Value"]

            req = "SELECT COUNT(JobID) FROM Jobs WHERE Status IN (%s) AND Site=%s" % (waitingString, e_site)
            result = self._query(req)
            if result["OK"]:
                count = result["Value"][0][0]
            else:
                return S_ERROR("Failed to get Site data from the JobDB")
            siteDict[site][JobStatus.WAITING] = count
            totalDict[JobStatus.WAITING] += count
            # Running,Stalled,Done,Failed
            for status in [
                '"%s"' % JobStatus.RUNNING,
                '"%s"' % JobStatus.STALLED,
                '"%s"' % JobStatus.DONE,
                '"%s"' % JobStatus.FAILED,
            ]:
                req = "SELECT COUNT(JobID) FROM Jobs WHERE Status=%s AND Site=%s" % (status, e_site)
                result = self._query(req)
                if result["OK"]:
                    count = result["Value"][0][0]
                else:
                    return S_ERROR("Failed to get Site data from the JobDB")
                siteDict[site][status.replace('"', "")] = count
                totalDict[status.replace('"', "")] += count

        siteDict["Total"] = totalDict
        return S_OK(siteDict)

    #################################################################################
    def getSiteSummaryWeb(self, selectDict, sortList, startItem, maxItems):
        """Get the summary of jobs in a given status on all the sites in the standard Web form"""

        paramNames = ["Site", "GridType", "Country", "Tier", "MaskStatus"]
        paramNames += JobStatus.JOB_STATES
        paramNames += ["Efficiency", "Status"]

        # Sort out records as requested
        sortItem = -1
        sortOrder = "ASC"
        if sortList:
            item = sortList[0][0]  # only one item for the moment
            sortItem = paramNames.index(item)
            sortOrder = sortList[0][1]

        last_update = None
        if "LastUpdateTime" in selectDict:
            last_update = selectDict["LastUpdateTime"]
            del selectDict["LastUpdateTime"]

        result = self.getCounters("Jobs", ["Site", "Status"], {}, newer=last_update, timeStamp="LastUpdateTime")
        last_day = Time.dateTime() - Time.day
        resultDay = self.getCounters("Jobs", ["Site", "Status"], {}, newer=last_day, timeStamp="EndExecTime")

        # Get the site mask status
        siteMask = {}
        resultMask = self.siteClient.getSites("All")
        if resultMask["OK"]:
            for site in resultMask["Value"]:
                siteMask[site] = "NoMask"
        resultMask = self.siteClient.getSites("Active")
        if resultMask["OK"]:
            for site in resultMask["Value"]:
                siteMask[site] = "Active"
        resultMask = self.siteClient.getSites("Banned")
        if resultMask["OK"]:
            for site in resultMask["Value"]:
                siteMask[site] = "Banned"

        # Sort out different counters
        resultDict = {}
        if result["OK"]:
            for attDict, count in result["Value"]:
                siteFullName = attDict["Site"]
                status = attDict["Status"]
                if siteFullName not in resultDict:
                    resultDict[siteFullName] = {}
                    for state in JobStatus.JOB_STATES:
                        resultDict[siteFullName][state] = 0
                if status not in JobStatus.JOB_FINAL_STATES:
                    resultDict[siteFullName][status] = count
        if resultDay["OK"]:
            for attDict, count in resultDay["Value"]:
                siteFullName = attDict["Site"]
                if siteFullName not in resultDict:
                    resultDict[siteFullName] = {}
                    for state in JobStatus.JOB_STATES:
                        resultDict[siteFullName][state] = 0
                status = attDict["Status"]
                if status in JobStatus.JOB_FINAL_STATES:
                    resultDict[siteFullName][status] = count

        # Collect records now
        records = []
        countryCounts = {}
        for siteFullName in resultDict:
            siteDict = resultDict[siteFullName]
            if siteFullName.count(".") == 2:
                grid, _, country = siteFullName.split(".")
            else:
                grid, _, country = "Unknown", "Unknown", "Unknown"

            res = getSiteTier(siteFullName)
            if not res["OK"]:
                self.log.error(res["Message"])
                continue
            tier = res["Value"]

            if country not in countryCounts:
                countryCounts[country] = {}
                for state in JobStatus.JOB_STATES:
                    countryCounts[country][state] = 0
            rList = [siteFullName, grid, country, tier]
            if siteFullName in siteMask:
                rList.append(siteMask[siteFullName])
            else:
                rList.append("NoMask")
            for status in JobStatus.JOB_STATES:
                rList.append(siteDict[status])
                countryCounts[country][status] += siteDict[status]
            efficiency = 0
            total_finished = 0
            for state in JobStatus.JOB_FINAL_STATES:
                total_finished += resultDict[siteFullName][state]
            if total_finished > 0:
                efficiency = float(siteDict[JobStatus.DONE] + siteDict[JobStatus.COMPLETED]) / float(total_finished)
            rList.append("%.1f" % (efficiency * 100.0))
            # Estimate the site verbose status
            if efficiency > 0.95:
                rList.append("Good")
            elif efficiency > 0.80:
                rList.append("Fair")
            elif efficiency > 0.60:
                rList.append("Poor")
            elif total_finished == 0:
                rList.append("Idle")
            else:
                rList.append("Bad")
            records.append(rList)

        # Select records as requested
        if selectDict:
            for item in selectDict:
                selectItem = paramNames.index(item)
                values = selectDict[item]
                if not isinstance(values, list):
                    values = [values]
                indices = list(range(len(records)))
                indices.reverse()
                for ind in indices:
                    if records[ind][selectItem] not in values:
                        del records[ind]

        # Sort records as requested
        if sortItem != -1:
            if sortOrder.lower() == "asc":
                records.sort(key=operator.itemgetter(sortItem))
            else:
                records.sort(key=operator.itemgetter(sortItem), reverse=True)

        # Collect the final result
        finalDict = {}
        finalDict["ParameterNames"] = paramNames
        # Return all the records if maxItems == 0 or the specified number otherwise
        if maxItems:
            if startItem + maxItems > len(records):
                finalDict["Records"] = records[startItem:]
            else:
                finalDict["Records"] = records[startItem : startItem + maxItems]
        else:
            finalDict["Records"] = records

        finalDict["TotalRecords"] = len(records)
        finalDict["Extras"] = countryCounts

        return S_OK(finalDict)

    #####################################################################################
    def setHeartBeatData(self, jobID, dynamicDataDict):
        """Add the job's heart beat data to the database"""

        # Set the time stamp first
        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        e_jobID = ret["Value"]

        req = "UPDATE Jobs SET HeartBeatTime=UTC_TIMESTAMP(), Status='%s' WHERE JobID=%s" % (JobStatus.RUNNING, e_jobID)
        result = self._update(req)
        if not result["OK"]:
            return S_ERROR("Failed to set the heart beat time: " + result["Message"])

        ok = True
        # Add dynamic data to the job heart beat log
        # start = time.time()
        valueList = []
        for key, value in dynamicDataDict.items():
            result = self._escapeString(key)
            if not result["OK"]:
                self.log.warn("Failed to escape string ", key)
                continue
            e_key = result["Value"]
            result = self._escapeString(value)
            if not result["OK"]:
                self.log.warn("Failed to escape string ", value)
                continue
            e_value = result["Value"]
            valueList.append("( %s, %s,%s,UTC_TIMESTAMP())" % (e_jobID, e_key, e_value))

        if valueList:

            valueString = ",".join(valueList)
            req = "INSERT INTO HeartBeatLoggingInfo (JobID,Name,Value,HeartBeatTime) VALUES "
            req += valueString
            result = self._update(req)
            if not result["OK"]:
                ok = False
                self.log.warn(result["Message"])

        if ok:
            return S_OK()
        return S_ERROR("Failed to store some or all the parameters")

    #####################################################################################
    def getHeartBeatData(self, jobID):
        """Retrieve the job's heart beat data"""
        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        jobID = ret["Value"]

        cmd = "SELECT Name,Value,HeartBeatTime from HeartBeatLoggingInfo WHERE JobID=%s" % jobID
        res = self._query(cmd)
        if not res["OK"]:
            return res

        if not res["Value"]:
            return S_OK([])

        result = []
        values = res["Value"]
        for row in values:
            name, value, heartbeattime = row
            if six.PY3 and isinstance(value, bytes):
                value = value.decode()
            result.append((str(name), "%.01f" % (float(value.replace('"', ""))), str(heartbeattime)))

        return S_OK(result)

    #####################################################################################
    def setJobCommand(self, jobID, command, arguments=None):
        """Store a command to be passed to the job together with the
        next heart beat
        """
        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        jobID = ret["Value"]

        ret = self._escapeString(command)
        if not ret["OK"]:
            return ret
        command = ret["Value"]

        if arguments:
            ret = self._escapeString(arguments)
            if not ret["OK"]:
                return ret
            arguments = ret["Value"]
        else:
            arguments = "''"

        req = "INSERT INTO JobCommands (JobID,Command,Arguments,ReceptionTime) "
        req += "VALUES (%s,%s,%s,UTC_TIMESTAMP())" % (jobID, command, arguments)
        return self._update(req)

    #####################################################################################
    def getJobCommand(self, jobID, status="Received"):
        """Get a command to be passed to the job together with the
        next heart beat
        """

        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        jobID = ret["Value"]

        ret = self._escapeString(status)
        if not ret["OK"]:
            return ret
        status = ret["Value"]

        req = "SELECT Command, Arguments FROM JobCommands WHERE JobID=%s AND Status=%s" % (jobID, status)
        result = self._query(req)
        if not result["OK"]:
            return result

        resultDict = {}
        if result["Value"]:
            for row in result["Value"]:
                resultDict[row[0]] = row[1]

        return S_OK(resultDict)

    #####################################################################################
    def setJobCommandStatus(self, jobID, command, status):
        """Set the command status"""
        ret = self._escapeString(jobID)
        if not ret["OK"]:
            return ret
        jobID = ret["Value"]

        ret = self._escapeString(command)
        if not ret["OK"]:
            return ret
        command = ret["Value"]

        ret = self._escapeString(status)
        if not ret["OK"]:
            return ret
        status = ret["Value"]

        req = "UPDATE JobCommands SET Status=%s WHERE JobID=%s AND Command=%s" % (status, jobID, command)
        return self._update(req)

    #####################################################################################
    def getSummarySnapshot(self, requestedFields=False):
        """Get the summary snapshot for a given combination"""
        if not requestedFields:
            requestedFields = ["Status", "MinorStatus", "Site", "Owner", "OwnerGroup", "JobGroup", "JobSplitType"]
        defFields = ["DIRACSetup"] + requestedFields
        valueFields = ["COUNT(JobID)", "SUM(RescheduleCounter)"]
        defString = ", ".join(defFields)
        valueString = ", ".join(valueFields)
        sqlCmd = "SELECT %s, %s From Jobs GROUP BY %s" % (defString, valueString, defString)
        result = self._query(sqlCmd)
        if not result["OK"]:
            return result
        return S_OK(((defFields + valueFields), result["Value"]))

    def removeInfoFromHeartBeatLogging(self, status, delTime, maxLines):
        """Remove HeartBeatLoggingInfo from DB.

        :param str status: status of the jobs
        :param str delTime: timestamp of the age of the jobs
        :param int maxLines: maximum number of lines to be removed
        :returns: S_OK/S_ERROR
        """
        ret = self._escapeString(status)
        if not ret["OK"]:
            return ret
        status = ret["Value"]

        ret = self._escapeString(delTime)
        if not ret["OK"]:
            return ret
        delTime = ret["Value"]

        self.log.verbose("Removing HeartBeatLogginInfo for", "%r %r %r" % (status, delTime, maxLines))
        cmd = """DELETE h FROM HeartBeatLoggingInfo AS h
             JOIN (SELECT hi.JobID FROM HeartBeatLoggingInfo AS hi
                LEFT JOIN Jobs j on j.JobID = hi.JobID
                WHERE j.Status = %(status)s
                    AND
                  LastUpdateTime < %(delay)s
                LIMIT %(maxLines)d) h2
              ON h2.JobID = h.JobID""" % {
            "maxLines": maxLines,
            "status": status,
            "delay": delTime,
        }
        result = self._update(cmd)
        self.log.verbose("Removed from HBLI", result)
        return result



"""
Resource Usage
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os

from DIRAC import gLogger


class ResourceUsage(object):
    """Resource Usage is an abstract class that has to be implemented for every batch system used by DIRAC
    to get the resource usage of a given job. This information can then be processed by other modules
    (e.g. getting the time left in a Pilot)
    """

    def __init__(self, batchSystemName, jobIdEnvVar):
        """Standard constructor"""
        self.log = gLogger.getSubLogger("%sResourceUsage" % batchSystemName)
        self.jobID = os.environ.get(jobIdEnvVar)

    def getResourceUsage(self):
        """Returns S_OK with a dictionary that can contain entries:

        - CPU: the CPU time consumed since the beginning of the execution for current slot (seconds)
        - CPULimit: the CPU time limit for current slot (seconds)
        - WallClock: the wall clock time consumed since the beginning of the execution for current slot (seconds)
        - WallClockLimit: the wall clock time limit for current slot (seconds)
        - Unit: indicates whether the Batch System allocates resources with limited CPU time and/or wallclock time
          Unit can take the following values: 'CPU', 'WallClock' or 'Both'.

        :return: dict such as {CPU, CPULimit, WallClock, WallClockLimit, Unit}
        """
        raise NotImplementedError("getResourceUsage not implemented")



""" The TimeLeft utility allows to calculate the amount of CPU time
    left for a given batch system slot.  This is essential for the 'Filling
    Mode' where several VO jobs may be executed in the same allocated slot.

    The prerequisites for the utility to run are:
      - Plugin for extracting information from local batch system
      - Scale factor for the local site.

    With this information the utility can calculate in normalized units the
    CPU time remaining for a given slot.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import shlex

import DIRAC

from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.Core.Utilities.Subprocess import systemCall


class TimeLeft(object):
    """This generally does not run alone"""

    def __init__(self):
        """Standard constructor"""
        self.log = gLogger.getSubLogger("TimeLeft")

        self.cpuPower = gConfig.getValue("/LocalSite/CPUNormalizationFactor", 0.0)
        if not self.cpuPower:
            self.log.warn("/LocalSite/CPUNormalizationFactor not defined for site %s" % DIRAC.siteName())

        result = self.__getBatchSystemPlugin()
        if result["OK"]:
            self.batchPlugin = result["Value"]
        else:
            self.batchPlugin = None
            self.batchError = result["Message"]

    def getScaledCPU(self, processors=1):
        """Returns the current CPU Time spend (according to batch system) scaled according
        to /LocalSite/CPUNormalizationFactor
        """
        # Quit if no scale factor available
        if not self.cpuPower:
            return 0

        # Quit if Plugin is not available
        if not self.batchPlugin:
            return 0

        resourceDict = self.batchPlugin.getResourceUsage()

        if "Value" in resourceDict:
            if resourceDict["Value"].get("CPU"):
                return resourceDict["Value"]["CPU"] * self.cpuPower
            elif resourceDict["Value"].get("WallClock"):
                # When CPU value missing, guess from WallClock and number of processors
                return resourceDict["Value"]["WallClock"] * self.cpuPower * processors

        return 0

    def getTimeLeft(self, cpuConsumed=0.0, processors=1):
        """Returns the CPU Time Left for supported batch systems.
        The CPUConsumed is the current raw total CPU.
        """
        # Quit if no norm factor available
        if not self.cpuPower:
            return S_ERROR("/LocalSite/CPUNormalizationFactor not defined for site %s" % DIRAC.siteName())

        if not self.batchPlugin:
            return S_ERROR(self.batchError)

        resourceDict = self.batchPlugin.getResourceUsage()
        if not resourceDict["OK"]:
            self.log.warn("Could not determine timeleft for batch system at site %s" % DIRAC.siteName())
            return resourceDict

        resources = resourceDict["Value"]
        self.log.debug("self.batchPlugin.getResourceUsage(): %s" % str(resources))
        if not resources.get("CPULimit") and not resources.get("WallClockLimit"):
            # This should never happen
            return S_ERROR("No CPU or WallClock limit obtained")

        # if one of CPULimit or WallClockLimit is missing, compute a reasonable value
        if not resources.get("CPULimit"):
            resources["CPULimit"] = resources["WallClockLimit"] * processors
        elif not resources.get("WallClockLimit"):
            resources["WallClockLimit"] = resources["CPULimit"] / processors

        # if one of CPU or WallClock is missing, compute a reasonable value
        if not resources.get("CPU"):
            resources["CPU"] = resources["WallClock"] * processors
        elif not resources.get("WallClock"):
            resources["WallClock"] = resources["CPU"] / processors

        cpu = float(resources["CPU"])
        cpuLimit = float(resources["CPULimit"])
        wallClock = float(resources["WallClock"])
        wallClockLimit = float(resources["WallClockLimit"])
        batchSystemTimeUnit = resources.get("Unit", "Both")

        # Some batch systems rely on wall clock time and/or cpu time to make allocations
        if batchSystemTimeUnit == "WallClock":
            time = wallClock
            timeLimit = wallClockLimit
        else:
            time = cpu
            timeLimit = cpuLimit

        if time and cpuConsumed > 3600.0 and self.cpuPower:
            # If there has been more than 1 hour of consumed CPU and
            # there is a Normalization set for the current CPU
            # use that value to renormalize the values returned by the batch system
            # NOTE: cpuConsumed is non-zero for call by the JobAgent and 0 for call by the watchdog
            # cpuLimit and cpu may be in the units of the batch system, not real seconds...
            # (in this case the other case won't work)
            # therefore renormalise it using cpuConsumed (which is in real seconds)
            cpuWorkLeft = (timeLimit - time) * self.cpuPower * cpuConsumed / time
        else:
            # FIXME: this is always used by the watchdog... Also used by the JobAgent
            #        if consumed less than 1 hour of CPU
            # It was using self.scaleFactor but this is inconsistent: use the same as above
            # In case the returned cpu and cpuLimit are not in real seconds, this is however rubbish
            cpuWorkLeft = (timeLimit - time) * self.cpuPower

        self.log.verbose("Remaining CPU in normalized units is: %.02f" % cpuWorkLeft)
        return S_OK(cpuWorkLeft)

    def __getBatchSystemPlugin(self):
        """Using the name of the batch system plugin, will return an instance of the plugin class."""
        batchSystems = {
            "LSF": "LSB_JOBID",
            "PBS": "PBS_JOBID",
            "BQS": "QSUB_REQNAME",
            "SGE": "SGE_TASK_ID",
            "SLURM": "SLURM_JOB_ID",
            "HTCondor": "_CONDOR_JOB_AD",
        }  # more to be added later
        name = None
        for batchSystem, envVar in batchSystems.items():
            if envVar in os.environ:
                name = batchSystem
                break

        if name is None and "MACHINEFEATURES" in os.environ and "JOBFEATURES" in os.environ:
            # Only use MJF if legacy batch system information not available for now
            name = "MJF"

        if name is None:
            self.log.warn("Batch system type for site %s is not currently supported" % DIRAC.siteName())
            return S_ERROR("Current batch system is not supported")

        self.log.debug("Creating plugin for %s batch system" % (name))
        try:
            batchSystemName = "%sResourceUsage" % (name)
            batchPlugin = __import__(
                "DIRAC.Resources.Computing.BatchSystems.TimeLeft.%s"
                % batchSystemName,  # pylint: disable=unused-variable
                globals(),
                locals(),
                [batchSystemName],
            )
        except ImportError as x:
            msg = "Could not import DIRAC.Resources.Computing.BatchSystems.TimeLeft.%s" % (batchSystemName)
            self.log.warn(x)
            self.log.warn(msg)
            return S_ERROR(msg)

        try:
            batchStr = "batchPlugin.%s()" % (batchSystemName)
            batchInstance = eval(batchStr)
        except Exception as x:  # pylint: disable=broad-except
            msg = "Could not instantiate %s()" % (batchSystemName)
            self.log.warn(x)
            self.log.warn(msg)
            return S_ERROR(msg)

        return S_OK(batchInstance)


#############################################################################


def runCommand(cmd, timeout=120):
    """Wrapper around systemCall to return S_OK(stdout) or S_ERROR(message)"""
    result = systemCall(timeout=timeout, cmdSeq=shlex.split(cmd))
    if not result["OK"]:
        return result
    status, stdout, stderr = result["Value"][0:3]

    if status:
        gLogger.warn("Status %s while executing %s" % (status, cmd))
        gLogger.warn(stderr)
        if stdout:
            return S_ERROR(stdout)
        if stderr:
            return S_ERROR(stderr)
        return S_ERROR("Status %s while executing %s" % (status, cmd))
    else:
        return S_OK(str(stdout))



########################################################################
# File: Graph.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2012/09/27 07:22:15
########################################################################
"""
:mod: Graph

.. module: Graph

:synopsis: graph

.. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

graph
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"
import six

# #
# @file Graph.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2012/09/27 07:22:23
# @brief Definition of Graph class.
# pylint: disable=no-member


class DynamicProps(type):
    """

    metaclass allowing to create properties on the fly
    """

    def __new__(cls, name, bases, classdict):
        """
        new operator
        """

        def makeProperty(self, name, value, readOnly=False):
            """
            Add property :name: to class

            This also creates a private :_name: attribute
            If you want to make read only property, set :readOnly: flag to True
            :warn: could raise AttributeError if :name: of :_name: is already
            defined as an attribute
            """
            if hasattr(self, "_" + name) or hasattr(self, name):
                raise AttributeError("_%s or %s is already defined as a member" % (name, name))

            def fget(self):
                return self._getProperty(name)

            fset = None if readOnly else lambda self, value: self._setProperty(name, value)
            setattr(self, "_" + name, value)
            setattr(self.__class__, name, property(fget=fget, fset=fset))

        def _setProperty(self, name, value):
            """
            property setter
            """
            setattr(self, "_" + name, value)

        def _getProperty(self, name):
            """
            property getter
            """
            return getattr(self, "_" + name)

        classdict["makeProperty"] = makeProperty
        classdict["_setProperty"] = _setProperty
        classdict["_getProperty"] = _getProperty
        return type.__new__(cls, name, bases, classdict)


@six.add_metaclass(DynamicProps)
class Node(object):
    """
    graph node
    """

    def __init__(self, name, rwAttrs=None, roAttrs=None):
        """
        c'tor

        :param str name: node name
        :param dict rwAttrs: read/write properties dict
        :param dict roAttrs: read-only properties dict
        """
        self.makeProperty("name", name, True)
        self.makeProperty("visited", False)
        self.__edges = list()
        rwAttrs = rwAttrs if isinstance(rwAttrs, dict) else {}
        for attr, value in rwAttrs.items():
            self.makeProperty(attr, value, False)
        roAttrs = roAttrs if isinstance(roAttrs, dict) else {}
        for attr, value in roAttrs.items():
            self.makeProperty(attr, value, True)

    def __contains__(self, edge):
        """
        in operator for edges
        """
        if not isinstance(edge, Edge):
            raise TypeError("edge should be an instance or subclass of Edge")
        return edge in self.__edges

    def __iter__(self):
        """
        edges iterator
        """
        return self.__edges.__iter__()

    def edges(self):
        """
        get edges
        """
        return self.__edges

    def addEdge(self, edge):
        """
        Add edge to the node
        """
        if not isinstance(edge, Edge):
            raise TypeError("supplied edge argument should be an Edge instance or subclass")
        if edge not in self:
            self.__edges.append(edge)

    def connect(self, other, rwAttrs=None, roAttrs=None):
        """
        Connect self to Node :other: with edge attibutes rw :rwAttrs: and
        ro :roAttrs:
        """
        if not isinstance(other, Node):
            raise TypeError("argument other should be a Node instance!")
        edge = Edge(self, other, rwAttrs, roAttrs)
        if edge not in self:
            self.__edges.append(edge)
        return edge


@six.add_metaclass(DynamicProps)
class Edge(object):
    """
    Directed link between two nodes
    """

    def __init__(self, fromNode, toNode, rwAttrs=None, roAttrs=None):
        """
        c'tor

        :param Node fromNode: edge start
        :param Node toNode: edge end
        :param dict rwAttrs: read/write properties dict
        :param dict roAttrs: read only properties dict
        """
        if not isinstance(fromNode, Node):
            raise TypeError("supplied argument fromNode should be a Node instance")
        if not isinstance(toNode, Node):
            raise TypeError("supplied argument toNode should be a Node instance")
        self.makeProperty("fromNode", fromNode, True)
        self.makeProperty("toNode", toNode, True)
        self.makeProperty("name", "%s-%s" % (self.fromNode.name, self.toNode.name), True)
        self.makeProperty("visited", False)
        rwAttrs = rwAttrs if isinstance(rwAttrs, dict) else {}
        for attr, value in rwAttrs.items():
            self.makeProperty(attr, value, False)
        roAttrs = roAttrs if isinstance(roAttrs, dict) else {}
        for attr, value in roAttrs.items():
            self.makeProperty(attr, value, True)
        if self not in self.fromNode:
            self.fromNode.addEdge(self)

    def __str__(self):
        """
        str representation of an object
        """
        return self.name

    def __repr__(self):
        """
        repr operator for dot format
        """
        return "'%s' -> '%s';" % (self.fromNode.name, self.toNode.name)


########################################################################


@six.add_metaclass(DynamicProps)
class Graph(object):
    """
    A generic directed graph with attributes attached to its nodes and edges
    """

    # # flag to indicate PREORDER traversal
    __PREORDER = False
    # # flag to indicate POSTORDER traversal
    __POSTORDER = False

    def __init__(self, name, nodes=None, edges=None):
        """
        c'tor

        :param self: self reference
        :param str name: graph name
        :param list nodes: initial node list
        :param list edges: initial edge list
        """
        self.makeProperty("name", name, True)
        nodes = nodes if nodes else list()
        edges = edges if edges else list()
        self.__nodes = []
        self.__edges = []
        for edge in edges:
            if edge not in self:
                self.addEdge(edge)
        for node in nodes:
            if node not in self:
                self.addNode(node)

    def __contains__(self, obj):
        """
        In operator for edges and nodes
        """
        return bool(obj in self.__nodes or obj in self.__edges)

    def nodes(self):
        """
        Get nodes dict
        """
        return self.__nodes

    def getNode(self, nodeName):
        """
        Get node :nodeName:
        """
        for node in self.__nodes:
            if node.name == nodeName:
                return node

    def edges(self):
        """
        Get edges dict
        """
        return self.__edges

    def getEdge(self, edgeName):
        """
        Get edge :edgeName:
        """
        for edge in self.__edges:
            if edge.name == edgeName:
                return edge

    @property
    def PREORDER(self):
        """
        PREORDER getter
        """
        return self.__PREORDER

    @PREORDER.setter
    def PREORDER(self, flag=True):
        """
        PREORDER setter
        """
        self.__PREORDER = bool(flag)
        self.__POSTORDER = not self.__PREORDER

    @property
    def POSTORDER(self):
        """
        POSTORDER getter
        """
        return self.__POSTORDER

    @POSTORDER.setter
    def POSTORDER(self, flag=True):
        """
        POSTORDER setter
        """
        self.__POSTORDER = bool(flag)
        self.__PREORDER = not self.__POSTORDER

    def connect(self, fromNode, toNode, rwAttrs=None, roAttrs=None):
        """
        Connect :fromNode: to :toNode: with edge of attributes
        """
        edge = fromNode.connect(toNode, rwAttrs, roAttrs)
        self.addEdge(edge)
        self.addNode(fromNode)
        self.addNode(toNode)
        return edge

    def addNode(self, node):
        """
        Add Node :node: to graph
        """
        if not isinstance(node, Node):
            raise TypeError("Supplied argument should be a Node instance")
        if node not in self:
            self.__nodes.append(node)
            if not hasattr(node, "graph"):
                node.makeProperty("graph", self)
            else:
                node.graph = self
        for edge in node:
            if edge not in self:
                self.addEdge(edge)
                if edge.toNode not in self:
                    self.addNode(edge.toNode)

    def addEdge(self, edge):
        """
        Add edge :edge: to the graph
        """
        if not isinstance(edge, Edge):
            raise TypeError("Supplied edge argument should be an Edge instance")
        if edge.fromNode not in self:
            self.addNode(edge.fromNode)
        if edge.toNode not in self:
            self.addNode(edge.toNode)
        if edge not in self:
            self.__edges.append(edge)
        if not hasattr(edge, "graph"):
            edge.makeProperty("graph", self)
        else:
            edge.graph = self

    def reset(self):
        """
        Set visited for all nodes to False
        """
        for node in self.__nodes:
            node.visited = False
        for edge in self.__edges:
            edge.visited = False

    def walkAll(self, nodeFcn=None, edgeFcn=None, res=None):
        """
        Wall all nodes excuting :nodeFcn: on each node and :edgeFcn: on each
        edge result is a dict { Node.name : result from :nodeFcn:,
        Edge.name : result from edgeFcn }
        """
        if not any((self.PREORDER, self.POSTORDER)):
            self.PREORDER = True
        res = res if res else {}
        self.reset()
        for node in self.nodes():
            if not node.visited:
                res.update(self.walkNode(node, nodeFcn, edgeFcn, res))
        return res

    def walkNode(self, node, nodeFcn=None, edgeFcn=None, res=None):
        """
        Walk through the graph calling nodeFcn on nodes and edgeFcn on edges
        """
        res = res if res else {}
        # # already visited, return
        if node.visited:
            return res

        if self.PREORDER:
            node.visited = True
            if callable(nodeFcn):
                res.update({node.name: nodeFcn(node)})
            for edge in node:
                # # execute edge fcn
                if callable(edgeFcn):
                    res[edge.name] = edgeFcn(edge)
                # # mark edge visited
                edge.visited = True
                res.update(self.walkNode(edge.toNode, nodeFcn, edgeFcn, res))
            return res

        if self.POSTORDER:
            node.visited = True
            for edge in node:
                # # execute edge fcn
                if callable(edgeFcn):
                    res[edge.name] = edgeFcn(edge)
                # # mark edge visited
                edge.visited = True
                res.update(self.walkNode(edge.toNode, nodeFcn, edgeFcn, res))
            if callable(nodeFcn):
                res.update({node.name: nodeFcn(node)})
            return res

    def __repr__(self):
        """
        Repr operator creating dot string
        """
        out = ["digraph '%s' {" % self.name]
        for node in self.nodes():
            out.append("%s;" % node.name)
        for edge in self.edges():
            out.append(repr(edge))
        out.append("}")
        return "\n".join(out)

    def explore(self, node, preVisit=None, postVisit=None):
        """
        Explore node
        """
        node.visited = True
        if callable(preVisit):
            preVisit(node)
        for edge in node.edges():
            if not edge.toNode.visited:
                self.explore(edge.toNode, preVisit, postVisit)
        if callable(postVisit):
            postVisit(node)

    def dfs(self, preVisit=None, postVisit=None):
        """
        dfs recursive walk
        """
        self.reset()
        nodes = list(self.nodes())
        nodes.sort(key=lambda node: len(node.edges()), reverse=True)
        for node in nodes:
            if not node.visited:
                self.explore(node, preVisit, postVisit)

    def bfs(self, preVisit=None, postVisit=None):
        """
        bfs walk
        """
        self.reset()
        nodes = list(self.nodes())
        nodes.sort(key=lambda node: len(node.edges()), reverse=True)
        queue = [nodes[0]]
        while queue:
            node = queue.pop(0)
            if callable(preVisit):
                preVisit(node)
            node.visited = True
            for edge in node.edges():
                if not edge.toNode.visited:
                    queue.append(edge.toNode)
            if callable(postVisit):
                postVisit(node)
        return nodes

    def dfsIter(self, preVisit=None, postVisit=None):
        """
        Iterative dfs - no recursion
        """
        nodes = list(self.nodes())
        nodes.sort(key=lambda node: len(node.edges()), reverse=True)
        for node in nodes:
            if not hasattr(node, "explored"):
                node.makeProperty("explored", False, False)
            if not hasattr(node, "discovered "):
                node.makeProperty("discovered", False, False)
        stack = []
        for node in nodes:
            if not node.discovered:
                node.discovered = True
                stack.append(node)
            while stack:
                node = stack[-1]
                if callable(preVisit):
                    preVisit(node)
                for edge in node.edges():
                    if not edge.toNode.discovered:
                        edge.toNode.discovered = True
                        stack.append(edge.toNode)
                if not node.explored:
                    node.explored = True
                if callable(postVisit):
                    postVisit(node)
                    stack.pop()



"""
Test LogRecord Creation

- createLogRecord
- debug, verbose, info, warn, notice, error, fatal, always

"""

__RCSID__ = "$Id$"

import pytest

from DIRAC.FrameworkSystem.private.standardLogging.LogLevels import LogLevels
from DIRAC.FrameworkSystem.private.standardLogging.test.TestLogUtilities import gLogger, gLoggerReset, cleaningLog


@pytest.mark.parametrize(
    "sMsg, sVarMsg, exc_info, expected",
    [
        ("message", "", False, "message"),
        ("this\nis\na\nmessage\non\nmultiple\nlines.", "", False, "this\nis\na\nmessage\non\nmultiple\nlines."),
        ("mess", "age", False, "mess age"),
        ("this\nis\na\nmessage\non\n", "multiple\nlines.", False, "this\nis\na\nmessage\non\n multiple\nlines."),
    ],
)
def test_createLogRecord(sMsg, sVarMsg, exc_info, expected):
    """
    Create logs of different levels with multiple logs
    """
    capturedBackend, log, sublog = gLoggerReset()

    # Set the level to debug
    gLogger.setLevel("debug")

    # dictionary of key = logger to use, value = output associated to the logger
    logDict = {gLogger: "", log: "/log", sublog: "/log/sublog"}

    # get list of existing levels, for each of them, a log record is created
    levels = gLogger.getAllPossibleLevels()
    for level in levels:
        for logger, logInfo in logDict.items():

            # createLogRecord is the method in charge of creating the log record
            # debug, ..., always methods wrap the following method
            # we use logLevels to get the int value corresponding to the level name
            logger._createLogRecord(LogLevels.getLevelValue(level), sMsg, sVarMsg, exc_info)

            # clean the log to remove unecessary information
            logstring = cleaningLog(capturedBackend.getvalue())
            logExpected = "Framework%s %s: %s\n" % (logInfo, level, expected)
            assert logExpected == logstring
            capturedBackend.truncate(0)
            capturedBackend.seek(0)


def test_showStack():
    """
    Get the showStack
    """
    capturedBackend, log, sublog = gLoggerReset()

    # dictionary of key = logger to use, value = output associated to the logger
    logDict = {gLogger: "", log: "/log", sublog: "/log/sublog"}
    for logger, logInfo in logDict.items():
        # By default, should not appear as the level is NOTICE
        logger.showStack()

        # clean the log to remove unecessary information
        logstring = cleaningLog(capturedBackend.getvalue())
        assert logstring == ""
        capturedBackend.truncate(0)
        capturedBackend.seek(0)

    # Set level to debug
    gLogger.setLevel("debug")

    for logger, logInfo in logDict.items():
        # The debug message should appear
        logger.showStack()

        # clean the log to remove unecessary information
        logstring = cleaningLog(capturedBackend.getvalue())
        assert logstring == "Framework%s DEBUG: \n" % logInfo
        capturedBackend.truncate(0)
        capturedBackend.seek(0)



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-get-job-pilots
# Author :  Stuart Paterson
########################################################################
"""
Retrieve info about pilots that have matched a given Job

Example:
  $ dirac-admin-get-job-pilots 1848
  {'https://marlb.in2p3.fr:9000/bqYViq6KrVgGfr6wwgT45Q': {'AccountingSent': 'False',
                                                          'BenchMark': 8.1799999999999997,
                                                          'Broker': 'marwms.in2p3.fr',
                                                          'DestinationSite': 'lpsc-ce.in2p3.fr',
                                                          'GridSite': 'LCG.LPSC.fr',
                                                          'GridType': 'gLite',
                                                          'Jobs': [1848L],
                                                          'LastUpdateTime': datetime.datetime(2011, 2, 21, 12, 39, 10),
                                                          'OutputReady': 'True',
                                                          'OwnerDN': '/O=GRID/C=FR/O=CNRS/OU=LPC/CN=Sebastien Guizard',
                                                          'OwnerGroup': '/biomed',
                                                          'ParentID': 0L,
                                                          'PilotID': 2247L,
                                                          'PilotJobReference': 'https://marlb.in2p3.fr:9000/biq6KT45Q',
                                                          'PilotStamp': '',
                                                          'Status': 'Done',
                                                          'SubmissionTime': datetime.datetime(2011, 2, 21, 12, 27, 52),
                                                          'TaskQueueID': 399L}}
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

# pylint: disable=wrong-import-position
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["JobID:    DIRAC ID of the Job"])
    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    _, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC import exit as DIRACExit
    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    exitCode = 0
    errorList = []

    for job in args:

        try:
            job = int(job)
        except Exception as x:
            errorList.append((job, "Expected integer for jobID"))
            exitCode = 2
            continue

        result = diracAdmin.getJobPilots(job)
        if not result["OK"]:
            errorList.append((job, result["Message"]))
            exitCode = 2

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRACExit(exitCode)


if __name__ == "__main__":
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from socket import socket, AF_INET, SOCK_DGRAM
import struct
import time as time
import datetime
from DIRAC import S_OK, S_ERROR

TIME1970 = 2208988800
gDefaultNTPServers = ["pool.ntp.org"]


def getNTPUTCTime(serverList=None, retries=2):
    data = "\x1b" + 47 * "\0"
    if not serverList:
        serverList = gDefaultNTPServers
    for server in serverList:
        client = socket(AF_INET, SOCK_DGRAM)
        client.settimeout(1)
        worked = False
        while retries >= 0 and not worked:
            try:
                client.sendto(data, (server, 123))
                data, address = client.recvfrom(1024)
                worked = True
            except Exception:
                retries -= 1
        if not worked:
            continue
        if data:
            myTime = struct.unpack("!12I", data)[10]
            myTime -= TIME1970
            return S_OK(datetime.datetime(*time.gmtime(myTime)[:6]))
    return S_ERROR("Could not get NTP time")


def getClockDeviation(serverList=None):
    result = getNTPUTCTime(serverList)
    if not result["OK"]:
        return result
    td = datetime.datetime.utcnow() - result["Value"]
    return S_OK(abs(td.days * 86400 + td.seconds))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Tornado.Server.TornadoService import TornadoService


def _check(result):
    assert result["OK"], result["Message"]
    return result["Value"]


def test_load():
    _check(ObjectLoader().loadObject("Core.Utilities.List", "fromChar"))
    _check(ObjectLoader().loadObject("Core.Utilities.ObjectLoader", "ObjectLoader"))
    assert _check(ObjectLoader().loadObject("Core.Utilities.ObjectLoader")) is ObjectLoader
    dataFilter = _check(ObjectLoader().getObjects("WorkloadManagementSystem.Service", ".*Handler"))
    dataClass = _check(ObjectLoader().getObjects("WorkloadManagementSystem.Service", parentClass=RequestHandler))
    tornadoDataClass = _check(ObjectLoader().getObjects("WorkloadManagementSystem.Service", parentClass=TornadoService))

    assert sorted(dataFilter) == sorted(set(dataClass).union(set(tornadoDataClass)))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import threading


class LockManager(object):
    def __init__(self, iMaxThreads=None):
        self.iMaxThreads = iMaxThreads
        if iMaxThreads:
            self.oGlobalLock = threading.Semaphore(iMaxThreads)
        else:
            self.oGlobalLock = False
        self.dLocks = {}
        self.dSubManagers = {}

    def createLock(self, sLockName, iMaxThreads):
        if sLockName in self.dLocks:
            raise RuntimeError("%s lock already exists" % sLockName)
        if iMaxThreads < 1:
            return
        self.dLocks[sLockName] = threading.Semaphore(iMaxThreads)
        self.dLocks[sLockName].release()

    def lockGlobal(self):
        if self.oGlobalLock:
            self.oGlobalLock.acquire()

    def unlockGlobal(self):
        if self.oGlobalLock:
            self.oGlobalLock.release()

    def lock(self, sLockName):
        if sLockName in self.dLocks:
            self.dLocks[sLockName].acquire()

    def unlock(self, sLockName):
        if sLockName in self.dLocks:
            self.dLocks[sLockName].release()



""" Encapsulate here the logic for limiting the matching of jobs

    Utilities and classes here are used by the Matcher
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id"

from DIRAC import S_OK, S_ERROR
from DIRAC import gLogger

from DIRAC.Core.Utilities.DictCache import DictCache
from DIRAC.Core.Utilities.DErrno import cmpError, ESECTION
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.WorkloadManagementSystem.DB.JobDB import JobDB
from DIRAC.WorkloadManagementSystem.Client import JobStatus


class Limiter(object):

    # static variables shared between all instances of this class
    csDictCache = DictCache()
    condCache = DictCache()
    delayMem = {}

    def __init__(self, jobDB=None, opsHelper=None, pilotRef=None):
        """Constructor"""
        self.__runningLimitSection = "JobScheduling/RunningLimit"
        self.__matchingDelaySection = "JobScheduling/MatchingDelay"

        if jobDB:
            self.jobDB = jobDB
        else:
            self.jobDB = JobDB()

        if pilotRef:
            self.log = gLogger.getSubLogger("[%s]Limiter" % pilotRef)
            self.jobDB.log = gLogger.getSubLogger("[%s]Limiter" % pilotRef)
        else:
            self.log = gLogger.getSubLogger("Limiter")

        if opsHelper:
            self.__opsHelper = opsHelper
        else:
            self.__opsHelper = Operations()

    def getNegativeCond(self):
        """Get negative condition for ALL sites"""
        orCond = self.condCache.get("GLOBAL")
        if orCond:
            return orCond
        negCond = {}
        # Run Limit
        result = self.__opsHelper.getSections(self.__runningLimitSection)
        sites = []
        if result["OK"]:
            sites = result["Value"]
        for siteName in sites:
            result = self.__getRunningCondition(siteName)
            if not result["OK"]:
                continue
            data = result["Value"]
            if data:
                negCond[siteName] = data
        # Delay limit
        result = self.__opsHelper.getSections(self.__matchingDelaySection)
        sites = []
        if result["OK"]:
            sites = result["Value"]
        for siteName in sites:
            result = self.__getDelayCondition(siteName)
            if not result["OK"]:
                continue
            data = result["Value"]
            if not data:
                continue
            if siteName in negCond:
                negCond[siteName] = self.__mergeCond(negCond[siteName], data)
            else:
                negCond[siteName] = data
        orCond = []
        for siteName in negCond:
            negCond[siteName]["Site"] = siteName
            orCond.append(negCond[siteName])
        self.condCache.add("GLOBAL", 10, orCond)
        return orCond

    def getNegativeCondForSite(self, siteName, gridCE=None):
        """Generate a negative query based on the limits set on the site"""
        # Check if Limits are imposed onto the site
        negativeCond = {}
        if self.__opsHelper.getValue("JobScheduling/CheckJobLimits", True):
            result = self.__getRunningCondition(siteName)
            if not result["OK"]:
                self.log.error("Issue getting running conditions", result["Message"])
            else:
                negativeCond = result["Value"]
            self.log.verbose(
                "Negative conditions for site", "%s after checking limits are: %s" % (siteName, str(negativeCond))
            )

            if gridCE:
                result = self.__getRunningCondition(siteName, gridCE)
                if not result["OK"]:
                    self.log.error("Issue getting running conditions", result["Message"])
                else:
                    negativeCondCE = result["Value"]
                    negativeCond = self.__mergeCond(negativeCond, negativeCondCE)

        if self.__opsHelper.getValue("JobScheduling/CheckMatchingDelay", True):
            result = self.__getDelayCondition(siteName)
            if result["OK"]:
                delayCond = result["Value"]
                self.log.verbose(
                    "Negative conditions for site", "%s after delay checking are: %s" % (siteName, str(delayCond))
                )
                negativeCond = self.__mergeCond(negativeCond, delayCond)

        if negativeCond:
            self.log.info("Negative conditions for site", "%s are: %s" % (siteName, str(negativeCond)))

        return negativeCond

    def __mergeCond(self, negCond, addCond):
        """Merge two negative dicts"""
        # Merge both negative dicts
        for attr in addCond:
            if attr not in negCond:
                negCond[attr] = []
            for value in addCond[attr]:
                if value not in negCond[attr]:
                    negCond[attr].append(value)
        return negCond

    def __extractCSData(self, section):
        """Extract limiting information from the CS in the form:
        { 'JobType' : { 'Merge' : 20, 'MCGen' : 1000 } }
        """
        stuffDict = self.csDictCache.get(section)
        if stuffDict:
            return S_OK(stuffDict)

        result = self.__opsHelper.getSections(section)
        if not result["OK"]:
            if cmpError(result, ESECTION):
                return S_OK({})
            return result
        attribs = result["Value"]
        stuffDict = {}
        for attName in attribs:
            result = self.__opsHelper.getOptionsDict("%s/%s" % (section, attName))
            if not result["OK"]:
                return result
            attLimits = result["Value"]
            try:
                attLimits = dict([(k, int(attLimits[k])) for k in attLimits])
            except Exception as excp:
                errMsg = "%s/%s has to contain numbers: %s" % (section, attName, str(excp))
                self.log.error(errMsg)
                return S_ERROR(errMsg)
            stuffDict[attName] = attLimits

        self.csDictCache.add(section, 300, stuffDict)
        return S_OK(stuffDict)

    def __getRunningCondition(self, siteName, gridCE=None):
        """Get extra conditions allowing site throttling"""
        if gridCE:
            csSection = "%s/%s/CEs/%s" % (self.__runningLimitSection, siteName, gridCE)
        else:
            csSection = "%s/%s" % (self.__runningLimitSection, siteName)
        result = self.__extractCSData(csSection)
        if not result["OK"]:
            return result
        limitsDict = result["Value"]
        # limitsDict is something like { 'JobType' : { 'Merge' : 20, 'MCGen' : 1000 } }
        if not limitsDict:
            return S_OK({})
        # Check if the site exceeding the given limits
        negCond = {}
        for attName in limitsDict:
            if attName not in self.jobDB.jobAttributeNames:
                self.log.error("Attribute does not exist", "(%s). Check the job limits" % attName)
                continue
            cK = "Running:%s:%s" % (siteName, attName)
            data = self.condCache.get(cK)
            if not data:
                result = self.jobDB.getCounters(
                    "Jobs",
                    [attName],
                    {"Site": siteName, "Status": [JobStatus.RUNNING, JobStatus.MATCHED, JobStatus.STALLED]},
                )
                if not result["OK"]:
                    return result
                data = result["Value"]
                data = dict([(k[0][attName], k[1]) for k in data])
                self.condCache.add(cK, 10, data)
            for attValue in limitsDict[attName]:
                limit = limitsDict[attName][attValue]
                running = data.get(attValue, 0)
                if running >= limit:
                    self.log.verbose(
                        "Job Limit imposed",
                        "at %s on %s/%s=%d, %d jobs already deployed" % (siteName, attName, attValue, limit, running),
                    )
                    if attName not in negCond:
                        negCond[attName] = []
                    negCond[attName].append(attValue)
        # negCond is something like : {'JobType': ['Merge']}
        return S_OK(negCond)

    def updateDelayCounters(self, siteName, jid):
        # Get the info from the CS
        siteSection = "%s/%s" % (self.__matchingDelaySection, siteName)
        result = self.__extractCSData(siteSection)
        if not result["OK"]:
            return result
        delayDict = result["Value"]
        # limitsDict is something like { 'JobType' : { 'Merge' : 20, 'MCGen' : 1000 } }
        if not delayDict:
            return S_OK()
        attNames = []
        for attName in delayDict:
            if attName not in self.jobDB.jobAttributeNames:
                self.log.error("Attribute does not exist in the JobDB. Please fix it!", "(%s)" % attName)
            else:
                attNames.append(attName)
        result = self.jobDB.getJobAttributes(jid, attNames)
        if not result["OK"]:
            self.log.error("Error while retrieving attributes", "coming from %s: %s" % (siteSection, result["Message"]))
            return result
        atts = result["Value"]
        # Create the DictCache if not there
        if siteName not in self.delayMem:
            self.delayMem[siteName] = DictCache()
        # Update the counters
        delayCounter = self.delayMem[siteName]
        for attName in atts:
            attValue = atts[attName]
            if attValue in delayDict[attName]:
                delayTime = delayDict[attName][attValue]
                self.log.notice("Adding delay for %s/%s=%s of %s secs" % (siteName, attName, attValue, delayTime))
                delayCounter.add((attName, attValue), delayTime)
        return S_OK()

    def __getDelayCondition(self, siteName):
        """Get extra conditions allowing matching delay"""
        if siteName not in self.delayMem:
            return S_OK({})
        lastRun = self.delayMem[siteName].getKeys()
        negCond = {}
        for attName, attValue in lastRun:
            if attName not in negCond:
                negCond[attName] = []
            negCond[attName].append(attValue)
        return S_OK(negCond)



""" a scheduler of threads, of course!
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import hashlib
import threading
import time

from DIRAC import S_ERROR, S_OK, gLogger
from DIRAC.Core.Utilities.ThreadSafe import Synchronizer

gSchedulerLock = Synchronizer()


class ThreadScheduler(object):
    def __init__(self, enableReactorThread=True, minPeriod=60):
        self.__thId = False
        self.__minPeriod = minPeriod
        self.__taskDict = {}
        self.__hood = []
        self.__createReactorThread = enableReactorThread
        self.__nowEpoch = time.time
        self.__sleeper = time.sleep
        self.__min = min

    def setMinValidPeriod(self, period):
        self.__minPeriod = period

    def disableCreateReactorThread(self):
        self.__createReactorThread = False

    def addPeriodicTask(self, period, taskFunc, taskArgs=(), executions=0, elapsedTime=0):
        if not callable(taskFunc):
            return S_ERROR("%s is not callable" % str(taskFunc))
        period = max(period, self.__minPeriod)
        elapsedTime = min(elapsedTime, period - 1)
        md = hashlib.md5()
        task = {
            "period": period,
            "func": taskFunc,
            "args": taskArgs,
        }
        md.update(str(task).encode())
        taskId = md.hexdigest()
        if taskId in self.__taskDict:
            return S_ERROR("Task %s is already added" % taskId)
        if executions:
            task["executions"] = executions
        self.__taskDict[taskId] = task
        retVal = self.__scheduleTask(taskId, elapsedTime)
        if not retVal["OK"]:
            return retVal
        self.__createExecutorIfNeeded()
        return S_OK(taskId)

    def setTaskPeriod(self, taskId, period):
        try:
            period = int(period)
        except ValueError:
            return S_ERROR("Period must be a number")
        period = max(period, self.__minPeriod)
        try:
            self.__taskDict[taskId]["period"] = period
        except KeyError:
            return S_ERROR("Unknown task %s" % taskId)
        return S_OK()

    @gSchedulerLock
    def removeTask(self, taskId):
        if taskId not in self.__taskDict:
            return S_ERROR("Task %s does not exist" % taskId)
        del self.__taskDict[taskId]
        for i in range(len(self.__hood)):
            if self.__hood[i][0] == taskId:
                del self.__hood[i]
                break
        return S_OK()

    def addSingleTask(self, taskFunc, taskArgs=()):
        return self.addPeriodicTask(self.__minPeriod, taskFunc, taskArgs, executions=1, elapsedTime=self.__minPeriod)

    @gSchedulerLock
    def __scheduleTask(self, taskId, elapsedTime=0):
        executeInSecs = self.__taskDict[taskId]["period"]
        elapsedTime = min(elapsedTime, executeInSecs - 1)
        if elapsedTime:
            executeInSecs -= elapsedTime

        now = time.time()
        for i in range(len(self.__hood)):
            if taskId == self.__hood[i][0]:
                tD = self.__hood[i][1] - now
                if abs(tD - executeInSecs) < 30:
                    return S_OK()
                else:
                    del self.__hood[i]
                    break

        executionTime = now + executeInSecs
        inserted = False
        for i in range(len(self.__hood)):
            if executionTime < self.__hood[i][1]:
                self.__hood.insert(i, (taskId, executionTime))
                inserted = True
                break
        if not inserted:
            self.__hood.append((taskId, executionTime))

        return S_OK()

    def __executorThread(self):
        while self.__hood:
            timeToNext = self.executeNextTask()
            if timeToNext is None:
                break
            if timeToNext and timeToNext > 0.1:
                self.__sleeper(self.__min(timeToNext, 1))
        # If we are leaving
        self.__destroyExecutor()

    def executeNextTask(self):
        if not self.__hood:
            return None
        timeToWait = self.__timeToNextTask()
        if timeToWait and timeToWait > 0:
            return timeToWait
        taskId = self.__popNextTaskId()
        startTime = time.time()
        self.__executeTask(taskId)
        elapsedTime = time.time() - startTime
        self.__scheduleIfNeeded(taskId, elapsedTime)
        return self.__timeToNextTask()

    @gSchedulerLock
    def __createExecutorIfNeeded(self):
        if not self.__createReactorThread:
            return
        if self.__thId:
            return
        self.__thId = threading.Thread(target=self.__executorThread)
        self.__thId.setDaemon(True)
        self.__thId.start()

    @gSchedulerLock
    def __destroyExecutor(self):
        self.__thId = False

    @gSchedulerLock
    def __timeToNextTask(self):
        if not self.__hood:
            return None
        return self.__hood[0][1] - self.__nowEpoch()

    @gSchedulerLock
    def __popNextTaskId(self):
        if len(self.__hood) == 0:
            return None
        return self.__hood.pop(0)[0]

    @gSchedulerLock
    def getNextTaskId(self):
        if len(self.__hood) == 0:
            return None
        return self.__hood[0][0]

    @gSchedulerLock
    def setNumExecutionsForTask(self, taskId, numExecutions):
        if taskId not in self.__taskDict:
            return False
        if numExecutions:
            self.__taskDict[taskId]["executions"] = numExecutions
        else:
            del self.__taskDict[taskId]["executions"]

    def __executeTask(self, taskId):
        if taskId not in self.__taskDict:
            return False
        task = self.__taskDict[taskId]
        if "executions" in task:
            task["executions"] -= 1
        try:
            task["func"](*task["args"])
        except Exception as lException:
            gLogger.exception("Exception while executing scheduled task", lException=lException)
            return False
        return True

    def __scheduleIfNeeded(self, taskId, elapsedTime=0):
        if "executions" in self.__taskDict[taskId]:
            if self.__taskDict[taskId]["executions"] == 0:
                del self.__taskDict[taskId]
                return True
        return self.__scheduleTask(taskId, elapsedTime)


gThreadScheduler = ThreadScheduler()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.AccountingSystem.Client.Types.BaseAccountingType import BaseAccountingType


class Pilot(BaseAccountingType):
    def __init__(self):
        super(Pilot, self).__init__()
        self.definitionKeyFields = [
            ("User", "VARCHAR(64)"),
            ("UserGroup", "VARCHAR(32)"),
            ("Site", "VARCHAR(64)"),
            ("GridCE", "VARCHAR(128)"),
            ("GridMiddleware", "VARCHAR(32)"),
            ("GridResourceBroker", "VARCHAR(128)"),
            ("GridStatus", "VARCHAR(32)"),
        ]
        self.definitionAccountingFields = [
            ("Jobs", "INT UNSIGNED"),
        ]
        self.checkType()
